from __future__ import annotations

import asyncio
import datetime
import json
import logging
import weakref
from collections import namedtuple
from typing import (
    Awaitable,
    Callable,
    Dict,
    Generic,
    Iterable,
    List,
    NoReturn,
    Optional,
    Set,
    Tuple,
    Union,
    overload,
)

import dask.config
import dask.distributed
from aiohttp import ContentTypeError
from dask.utils import parse_timedelta
from distributed.utils import Log, Logs
from rich.progress import Progress
from typing_extensions import TypeAlias

from coiled.cli.setup.entry import do_setup_wizard
from coiled.context import track_context
from coiled.core import Async, IsAsynchronous, Sync, delete_docstring, list_docstring
from coiled.core import Cloud as OldCloud
from coiled.errors import ClusterCreationError, DoesNotExist, ServerError
from coiled.types import (
    ArchitectureTypesEnum,
    AWSOptions,
    GCPOptions,
    PackageLevel,
    PackageSchema,
    ResolvedPackageInfo,
    SoftwareEnvironmentAlias,
)
from coiled.utils import (
    COILED_LOGGER_NAME,
    GatewaySecurity,
    get_grafana_url,
    validate_backend_options,
)

from .states import (
    InstanceStateEnum,
    ProcessStateEnum,
    flatten_log_states,
    get_process_instance_state,
    log_states,
)
from .widgets.util import simple_progress

logger = logging.getLogger(COILED_LOGGER_NAME)


def setup_logging(level=logging.INFO):
    # We want to be able to give info-level messages to users.
    # For users who haven't set up a log handler, this requires creating one (b/c the handler of "last resort,
    # logging.lastResort, has a level of "warning".
    #
    # Conservatively, we only do anything here if the user hasn't set up any log handlers on the root logger
    # or the Coiled logger. If they have any handler, we assume logging is configured how they want it.
    coiled_logger = logging.getLogger(COILED_LOGGER_NAME)
    root_logger = logging.getLogger()
    if coiled_logger.handlers == [] and root_logger.handlers == []:
        stream_handler = logging.StreamHandler()
        stream_handler.setFormatter(logging.Formatter(fmt="[%(asctime)s][%(levelname)-8s][%(name)s] %(message)s"))
        # Conservatively, only change the Coiled logger level there's no log level specified yet.
        if coiled_logger.level == 0:
            coiled_logger.setLevel(level)
            coiled_logger.addHandler(stream_handler)


async def handle_api_exception(response, exception_cls=ServerError) -> NoReturn:
    try:
        error_body = await response.json()
    except ContentTypeError:
        raise exception_cls(
            f"Unexpected status code ({response.status}) to {response.method}:{response.url}, contact support@coiled.io"
        ) from None
    if "message" in error_body:
        raise exception_cls(error_body["message"])
    if "detail" in error_body:
        raise exception_cls(error_body["detail"])
    raise exception_cls(error_body)


CloudV2SyncAsync: TypeAlias = Union["CloudV2[Async]", "CloudV2[Sync]"]


class CloudV2(OldCloud, Generic[IsAsynchronous]):
    _recent_sync: List[weakref.ReferenceType[CloudV2[Sync]]] = list()
    _recent_async: List[weakref.ReferenceType[CloudV2[Async]]] = list()

    # just overriding to get the right signature (CloudV2, not Cloud)
    def __enter__(self: CloudV2[Sync]) -> CloudV2[Sync]:
        return self

    def __exit__(self: CloudV2[Sync], typ, value, tb) -> None:
        self.close()

    async def __aenter__(self: CloudV2[Async]) -> CloudV2[Async]:
        return await self._start()

    async def __aexit__(self: CloudV2[Async], typ, value, tb) -> None:
        await self._close()

    # these overloads are necessary for the typechecker to know that we really have a CloudV2, not a Cloud
    # without them, CloudV2.current would be typed to return a Cloud
    #
    # https://www.python.org/dev/peps/pep-0673/ would remove the need for this.
    # That PEP also mentions a workaround with type vars, which doesn't work for us because type vars aren't
    # subscribtable
    @overload
    @classmethod
    def current(cls, asynchronous: Sync) -> CloudV2[Sync]: ...

    @overload
    @classmethod
    def current(cls, asynchronous: Async) -> CloudV2[Async]: ...

    @overload
    @classmethod
    def current(cls, asynchronous: bool) -> CloudV2: ...

    @classmethod
    def current(cls, asynchronous: bool) -> CloudV2:
        recent: List[weakref.ReferenceType[CloudV2]]
        if asynchronous:
            recent = cls._recent_async
        else:
            recent = cls._recent_sync
        try:
            cloud = recent[-1]()
            while cloud is None or cloud.status != "running":
                recent.pop()
                cloud = recent[-1]()
        except IndexError:
            if asynchronous:
                return cls(asynchronous=True)
            else:
                return cls(asynchronous=False)
        else:
            return cloud

    @track_context
    async def _get_default_instance_types(self, provider: str, gpu: bool = False, arch: str = "x86_64") -> List[str]:
        if arch not in ("arm64", "x86_64"):
            raise ValueError(f"arch '{arch}' is not supported for default instance types")
        if provider == "aws":
            if arch == "arm64":
                if gpu:
                    return ["g5g.xlarge"]  # has NVIDIA T4G
                else:
                    return ["m7g.xlarge", "m6g.xlarge"]
            if gpu:
                return ["g4dn.xlarge"]
            else:
                return ["m6i.xlarge", "m5.xlarge"]
        elif provider == "gcp":
            if arch != "x86_64":
                return ["t2a-standard-4"]
            if gpu:
                # n1-standard-8 with 30GB of memory might be best, but that's big for a default
                return ["n1-standard-4"]
            else:
                return ["e2-standard-4"]
        elif provider == "azure":
            if arch != "x86_64":
                raise ValueError(f"no default instance type for Azure with {arch} architecture")
            if gpu:
                raise ValueError("no default GPU instance type for Azure")
            return ["Standard_D4_v5"]
        else:
            raise ValueError(f"unexpected provider {provider}; cannot determine default instance types")

    async def _list_dask_scheduler_page(
        self,
        page: int,
        workspace: Optional[str] = None,
        since: Optional[str] = "7 days",
        user: Optional[str] = None,
    ) -> Tuple[list, bool]:
        page_size = 100
        workspace = workspace or self.default_workspace
        kwargs = {}
        if since:
            kwargs["since"] = parse_timedelta(since)
        if user:
            kwargs["user"] = user
        response = await self._do_request(
            "GET",
            self.server + f"/api/v2/analytics/{workspace}/clusters/list",
            params={
                "limit": page_size,
                "offset": page_size * page,
                **kwargs,
            },
        )
        if response.status >= 400:
            await handle_api_exception(response)

        results = await response.json()
        has_more_pages = len(results) > 0
        return results, has_more_pages

    @track_context
    async def _list_dask_scheduler(
        self,
        workspace: Optional[str] = None,
        since: Optional[str] = "7 days",
        user: Optional[str] = None,
    ):
        return await self._depaginate_list(
            self._list_dask_scheduler_page,
            workspace=workspace,
            since=since,
            user=user,
        )

    @overload
    def list_dask_scheduler(
        self: Cloud[Sync],
        account: Optional[str] = None,
        workspace: Optional[str] = None,
        since: Optional[str] = "7 days",
        user: Optional[str] = None,
    ) -> list: ...

    @overload
    def list_dask_scheduler(
        self: Cloud[Async],
        account: Optional[str] = None,
        workspace: Optional[str] = None,
        since: Optional[str] = "7 days",
        user: Optional[str] = "",
    ) -> Awaitable[list]: ...

    def list_dask_scheduler(
        self,
        account: Optional[str] = None,
        workspace: Optional[str] = None,
        since: Optional[str] = "7 days",
        user: Optional[str] = "",
    ) -> Union[list, Awaitable[list]]:
        return self._sync(self._list_dask_scheduler, workspace or account, since=since, user=user)

    async def _list_computations(
        self, cluster_id: Optional[int] = None, scheduler_id: Optional[int] = None, workspace: Optional[str] = None
    ):
        return await self._depaginate_list(
            self._list_computations_page, cluster_id=cluster_id, scheduler_id=scheduler_id, workspace=workspace
        )

    async def _list_computations_page(
        self,
        page: int,
        cluster_id: Optional[int] = None,
        scheduler_id: Optional[int] = None,
        workspace: Optional[str] = None,
    ) -> Tuple[list, bool]:
        page_size = 100
        workspace = workspace or self.default_workspace

        if not scheduler_id and not cluster_id:
            raise ValueError("either cluster_id or scheduler_id must be specified")

        api = (
            f"/api/v2/analytics/{workspace}/{scheduler_id}/computations/list"
            if scheduler_id
            else f"/api/v2/analytics/{workspace}/cluster/{cluster_id}/computations/list"
        )

        response = await self._do_request(
            "GET",
            self.server + api,
            params={"limit": page_size, "offset": page_size * page},
        )
        if response.status >= 400:
            await handle_api_exception(response)

        results = await response.json()
        has_more_pages = len(results) > 0
        return results, has_more_pages

    @overload
    def list_computations(
        self: Cloud[Sync],
        cluster_id: Optional[int] = None,
        scheduler_id: Optional[int] = None,
        account: Optional[str] = None,
        workspace: Optional[str] = None,
    ) -> list: ...

    @overload
    def list_computations(
        self: Cloud[Async],
        cluster_id: Optional[int] = None,
        scheduler_id: Optional[int] = None,
        account: Optional[str] = None,
        workspace: Optional[str] = None,
    ) -> Awaitable[list]: ...

    def list_computations(
        self,
        cluster_id: Optional[int] = None,
        scheduler_id: Optional[int] = None,
        account: Optional[str] = None,
        workspace: Optional[str] = None,
    ) -> Union[list, Awaitable[list]]:
        return self._sync(
            self._list_computations, cluster_id=cluster_id, scheduler_id=scheduler_id, workspace=workspace or account
        )

    def list_exceptions(
        self,
        cluster_id: Optional[int] = None,
        scheduler_id: Optional[int] = None,
        account: Optional[str] = None,
        workspace: Optional[str] = None,
        since: Optional[str] = None,
        user: Optional[str] = None,
    ) -> Union[list, Awaitable[list]]:
        return self._sync(
            self._list_exceptions,
            cluster_id=cluster_id,
            scheduler_id=scheduler_id,
            workspace=workspace or account,
            since=since,
            user=user,
        )

    async def _list_exceptions(
        self,
        cluster_id: Optional[int] = None,
        scheduler_id: Optional[int] = None,
        workspace: Optional[str] = None,
        since: Optional[str] = None,
        user: Optional[str] = None,
    ):
        return await self._depaginate_list(
            self._list_exceptions_page,
            cluster_id=cluster_id,
            scheduler_id=scheduler_id,
            workspace=workspace,
            since=since,
            user=user,
        )

    async def _list_exceptions_page(
        self,
        page: int,
        cluster_id: Optional[int] = None,
        scheduler_id: Optional[int] = None,
        workspace: Optional[str] = None,
        since: Optional[str] = None,
        user: Optional[str] = None,
    ) -> Tuple[list, bool]:
        page_size = 100
        workspace = workspace or self.default_workspace
        kwargs = {}
        if since:
            kwargs["since"] = parse_timedelta(since)
        if user:
            kwargs["user"] = user
        if cluster_id:
            kwargs["cluster"] = cluster_id
        if scheduler_id:
            kwargs["scheduler"] = scheduler_id
        response = await self._do_request(
            "GET",
            self.server + f"/api/v2/analytics/{workspace}/exceptions/list",
            params={"limit": page_size, "offset": page_size * page, **kwargs},
        )
        if response.status >= 400:
            await handle_api_exception(response)

        results = await response.json()
        has_more_pages = len(results) > 0
        return results, has_more_pages

    async def _list_events_page(
        self,
        page: int,
        cluster_id: int,
        workspace: Optional[str] = None,
    ) -> Tuple[list, bool]:
        page_size = 100
        workspace = workspace or self.default_workspace
        response = await self._do_request(
            "GET",
            self.server + f"/api/v2/analytics/{workspace}/{cluster_id}/events/list",
            params={"limit": page_size, "offset": page_size * page},
        )
        if response.status >= 400:
            await handle_api_exception(response)

        results = await response.json()
        has_more_pages = len(results) > 0
        return results, has_more_pages

    async def _list_events(self, cluster_id: int, workspace: Optional[str] = None):
        return await self._depaginate_list(self._list_events_page, cluster_id=cluster_id, workspace=workspace)

    def list_events(
        self,
        cluster_id: int,
        account: Optional[str] = None,
        workspace: Optional[str] = None,
    ) -> Union[list, Awaitable[list]]:
        return self._sync(self._list_events, cluster_id, workspace or account)

    async def _send_state(self, cluster_id: int, desired_status: str, workspace: Optional[str] = None):
        workspace = workspace or self.default_workspace
        response = await self._do_request(
            "POST",
            self.server + f"/api/v2/analytics/{workspace}/{cluster_id}/desired-state",
            json={"desired_status": desired_status},
        )
        if response.status >= 400:
            await handle_api_exception(response)

    def send_state(
        self,
        cluster_id: int,
        desired_status: str,
        account: Optional[str] = None,
        workspace: Optional[str] = None,
    ) -> Union[None, Awaitable[None]]:
        return self._sync(self._send_state, cluster_id, desired_status, workspace or account)

    @track_context
    async def _list_clusters(
        self, workspace: Optional[str] = None, max_pages: Optional[int] = None, just_mine: bool = False
    ):
        return await self._depaginate_list(
            self._list_clusters_page, workspace=workspace, max_pages=max_pages, just_mine=just_mine
        )

    @overload
    def list_clusters(
        self: Cloud[Sync],
        account: Optional[str] = None,
        workspace: Optional[str] = None,
        max_pages: Optional[int] = None,
        just_mine: bool = False,
    ) -> list: ...

    @overload
    def list_clusters(
        self: Cloud[Async],
        account: Optional[str] = None,
        workspace: Optional[str] = None,
        max_pages: Optional[int] = None,
        just_mine: bool = False,
    ) -> Awaitable[list]: ...

    @list_docstring
    def list_clusters(
        self,
        account: Optional[str] = None,
        workspace: Optional[str] = None,
        max_pages: Optional[int] = None,
        just_mine: bool = False,
    ) -> Union[list, Awaitable[list]]:
        return self._sync(self._list_clusters, workspace=workspace or account, max_pages=max_pages, just_mine=just_mine)

    async def _list_clusters_page(
        self, page: int, workspace: Optional[str] = None, just_mine: bool = False
    ) -> Tuple[list, bool]:
        page_size = 100
        workspace = workspace or self.default_workspace
        response = await self._do_request(
            "GET",
            self.server + f"/api/v2/clusters/account/{workspace}/",
            params={"limit": page_size, "offset": page_size * page, "just_mine": "1" if just_mine else "0"},
        )
        if response.status >= 400:
            await handle_api_exception(response)

        results = await response.json()
        has_more_pages = len(results) > 0
        return results, has_more_pages

    @staticmethod
    async def _depaginate_list(
        func: Callable[..., Awaitable[Tuple[list, bool]]],
        max_pages: Optional[int] = None,
        *args,
        **kwargs,
    ) -> list:
        results_all = []
        page = 0
        while True:
            kwargs["page"] = page
            results, next = await func(*args, **kwargs)
            results_all += results
            page += 1
            if (not results) or next is None:
                break
            # page is the number of pages we've already fetched (since 0-indexed)
            if max_pages and page >= max_pages:
                break
        return results_all

    @track_context
    async def _create_package_sync_env(
        self,
        packages: List[ResolvedPackageInfo],
        progress: Optional[Progress] = None,
        workspace: Optional[str] = None,
        gpu_enabled: bool = False,
        architecture: ArchitectureTypesEnum = ArchitectureTypesEnum.X86_64,
        region_name: Optional[str] = None,
        use_uv_installer: bool = True,
    ) -> SoftwareEnvironmentAlias:
        workspace = workspace or self.default_workspace
        prepared_packages: List[PackageSchema] = []
        for pkg in packages:
            if pkg["sdist"] and pkg["md5"]:
                with simple_progress(f"Uploading {pkg['name']}", progress=progress):
                    file_id = await self._create_senv_package(
                        pkg["sdist"],
                        contents_md5=pkg["md5"],
                        workspace=workspace,
                        region_name=region_name,
                    )
            else:
                file_id = None
            prepared_packages.append({
                "name": pkg["name"],
                "source": pkg["source"],
                "channel": pkg["channel"],
                "conda_name": pkg["conda_name"],
                "specifier": pkg["specifier"],
                "include": pkg["include"],
                "client_version": pkg["client_version"],
                "file": file_id,
            })
        with simple_progress("Requesting package sync build", progress=progress):
            result = await self._create_software_environment_v2(
                senv={
                    "packages": prepared_packages,
                    "raw_pip": None,
                    "raw_conda": None,
                },
                workspace=workspace,
                architecture=architecture,
                gpu_enabled=gpu_enabled,
                region_name=region_name,
                use_uv_installer=use_uv_installer,
            )
        return result

    @track_context
    async def _create_custom_certificate(self, subdomain: str, workspace: Optional[str] = None):
        workspace = workspace or self.default_workspace
        response = await self._do_request(
            "POST",
            self.server + f"/api/v2/clusters/account/{workspace}/https-certificate",
            json={"subdomain": subdomain},
        )
        if response.status >= 400:
            await handle_api_exception(response)

    async def _check_custom_certificate(self, subdomain: str, workspace: Optional[str] = None):
        response = await self._do_request(
            "GET",
            self.server + f"/api/v2/clusters/account/{workspace}/https-certificate/{subdomain}",
        )
        if response.status >= 400:
            await handle_api_exception(response)
        response_json = await response.json()
        cert_status = response_json.get("status")
        return cert_status

    @track_context
    async def _create_cluster(
        self,
        # todo: make name optional and pick one for them, like pre-declarative?
        # https://gitlab.com/coiled/cloud/-/issues/4305
        name: str,
        *,
        software_environment: Optional[str] = None,
        senv_v2_id: Optional[int] = None,
        worker_class: Optional[str] = None,
        worker_options: Optional[dict] = None,
        scheduler_options: Optional[dict] = None,
        workspace: Optional[str] = None,
        workers: int = 0,
        environ: Optional[Dict] = None,
        tags: Optional[Dict] = None,
        dask_config: Optional[Dict] = None,
        scheduler_vm_types: Optional[list] = None,
        gcp_worker_gpu_type: Optional[str] = None,
        gcp_worker_gpu_count: Optional[int] = None,
        worker_vm_types: Optional[list] = None,
        worker_disk_size: Optional[int] = None,
        worker_disk_throughput: Optional[int] = None,
        scheduler_disk_size: Optional[int] = None,
        backend_options: Optional[Union[AWSOptions, GCPOptions, dict]] = None,
        use_scheduler_public_ip: Optional[bool] = None,
        use_dashboard_https: Optional[bool] = None,
        private_to_creator: Optional[bool] = None,
        extra_worker_on_scheduler: Optional[bool] = None,
        n_worker_specs_per_host: Optional[int] = None,
        custom_subdomain: Optional[str] = None,
    ) -> Tuple[int, bool]:
        # TODO (Declarative): support these args, or decide not to
        # https://gitlab.com/coiled/cloud/-/issues/4305

        workspace = workspace or self.default_workspace
        account, name = self._normalize_name(name, context_workspace=workspace, allow_uppercase=True)

        await self._verify_workspace(account)

        data = {
            "name": name,
            "workers": workers,
            "worker_instance_types": worker_vm_types,
            "scheduler_instance_types": scheduler_vm_types,
            "worker_options": worker_options,
            "worker_class": worker_class,
            "worker_disk_size": worker_disk_size,
            "worker_disk_throughput": worker_disk_throughput,
            "scheduler_disk_size": scheduler_disk_size,
            "scheduler_options": scheduler_options,
            "environ": environ,
            "tags": tags,
            "dask_config": dask_config,
            "private_to_creator": private_to_creator,
            "env_id": senv_v2_id,
            "env_name": software_environment,
            "extra_worker_on_scheduler": extra_worker_on_scheduler,
            "n_worker_specs_per_host": n_worker_specs_per_host,
            "use_aws_creds_endpoint": dask.config.get("coiled.use_aws_creds_endpoint", False),
            "custom_subdomain": custom_subdomain,
        }

        backend_options = backend_options if backend_options else {}

        if gcp_worker_gpu_type is not None:
            # for backwards compatibility with v1 options
            backend_options = {
                **backend_options,
                "worker_accelerator_count": gcp_worker_gpu_count or 1,
                "worker_accelerator_type": gcp_worker_gpu_type,
            }
        elif gcp_worker_gpu_count:
            # not ideal but v1 only supported T4 and `worker_gpu=1` would give you one
            backend_options = {
                **backend_options,
                "worker_accelerator_count": gcp_worker_gpu_count,
                "worker_accelerator_type": "nvidia-tesla-t4",
            }

        if use_scheduler_public_ip is False:
            if "use_dashboard_public_ip" not in backend_options and not use_dashboard_https:
                backend_options["use_dashboard_public_ip"] = False

        if use_dashboard_https is False:
            if "use_dashboard_https" not in backend_options:
                backend_options["use_dashboard_https"] = False

        if backend_options:
            # for backwards compatibility with v1 options
            if "region" in backend_options and "region_name" not in backend_options:
                backend_options["region_name"] = backend_options["region"]  # type: ignore
                del backend_options["region"]  # type: ignore
            if "zone" in backend_options and "zone_name" not in backend_options:
                backend_options["zone_name"] = backend_options["zone"]  # type: ignore
                del backend_options["zone"]  # type: ignore
            # firewall just lets you specify a single CIDR block to open for ingress
            # we want to support a list of ingress CIDR blocks
            if "firewall" in backend_options:
                backend_options["ingress"] = [backend_options.pop("firewall")]  # type: ignore

            # convert the list of ingress rules to the FirewallSpec expected server-side
            if "ingress" in backend_options:
                fw_spec = {"ingress": backend_options.pop("ingress")}
                backend_options["firewall_spec"] = fw_spec  # type: ignore

            validate_backend_options(backend_options)
            data["options"] = backend_options

        response = await self._do_request(
            "POST",
            self.server + f"/api/v2/clusters/account/{account}/",
            json=data,
        )

        response_json = await response.json()

        if response.status >= 400:
            from .widgets import EXECUTION_CONTEXT

            if response_json.get("code") == "NO_CLOUD_SETUP":
                server_error_message = response_json.get("message")
                error_message = f"{server_error_message} or by running `coiled setup`"

                if EXECUTION_CONTEXT == "terminal":
                    # maybe not interactive so just raise
                    raise ClusterCreationError(error_message)
                else:
                    # interactive session so let's try running the cloud setup wizard
                    if await do_setup_wizard():
                        # the user setup their cloud backend, so let's try creating cluster again!
                        response = await self._do_request(
                            "POST",
                            self.server + f"/api/v2/clusters/account/{account}/",
                            json=data,
                        )
                        if response.status >= 400:
                            await handle_api_exception(response)  # always raises exception, no return
                        response_json = await response.json()
                    else:
                        raise ClusterCreationError(error_message)
            else:
                if "message" in response_json:
                    raise ServerError(response_json["message"])
                if "detail" in response_json:
                    raise ServerError(response_json["detail"])
                raise ServerError(response_json)

        return response_json["id"], response_json["existing"]

    @overload
    def create_cluster(
        self: Cloud[Sync],
        name: str,
        *,
        software: Optional[str] = None,
        worker_class: Optional[str] = None,
        worker_options: Optional[dict] = None,
        scheduler_options: Optional[dict] = None,
        account: Optional[str] = None,
        workspace: Optional[str] = None,
        workers: int = 0,
        environ: Optional[Dict] = None,
        tags: Optional[Dict] = None,
        dask_config: Optional[Dict] = None,
        private_to_creator: Optional[bool] = None,
        scheduler_vm_types: Optional[list] = None,
        worker_gpu_type: Optional[str] = None,
        worker_vm_types: Optional[list] = None,
        worker_disk_size: Optional[int] = None,
        worker_disk_throughput: Optional[int] = None,
        scheduler_disk_size: Optional[int] = None,
        backend_options: Optional[Union[dict, AWSOptions, GCPOptions]] = None,
    ) -> Tuple[int, bool]: ...

    @overload
    def create_cluster(
        self: Cloud[Async],
        name: str,
        *,
        software: Optional[str] = None,
        worker_class: Optional[str] = None,
        worker_options: Optional[dict] = None,
        scheduler_options: Optional[dict] = None,
        account: Optional[str] = None,
        workspace: Optional[str] = None,
        workers: int = 0,
        environ: Optional[Dict] = None,
        tags: Optional[Dict] = None,
        dask_config: Optional[Dict] = None,
        private_to_creator: Optional[bool] = None,
        scheduler_vm_types: Optional[list] = None,
        worker_gpu_type: Optional[str] = None,
        worker_vm_types: Optional[list] = None,
        worker_disk_size: Optional[int] = None,
        worker_disk_throughput: Optional[int] = None,
        scheduler_disk_size: Optional[int] = None,
        backend_options: Optional[Union[dict, AWSOptions, GCPOptions]] = None,
    ) -> Awaitable[Tuple[int, bool]]: ...

    def create_cluster(
        self,
        name: str,
        *,
        software: Optional[str] = None,
        worker_class: Optional[str] = None,
        worker_options: Optional[dict] = None,
        scheduler_options: Optional[dict] = None,
        account: Optional[str] = None,
        workspace: Optional[str] = None,
        workers: int = 0,
        environ: Optional[Dict] = None,
        tags: Optional[Dict] = None,
        private_to_creator: Optional[bool] = None,
        dask_config: Optional[Dict] = None,
        scheduler_vm_types: Optional[list] = None,
        worker_gpu_type: Optional[str] = None,
        worker_vm_types: Optional[list] = None,
        worker_disk_size: Optional[int] = None,
        worker_disk_throughput: Optional[int] = None,
        scheduler_disk_size: Optional[int] = None,
        backend_options: Optional[Union[dict, AWSOptions, GCPOptions]] = None,
    ) -> Union[Tuple[int, bool], Awaitable[Tuple[int, bool]]]:
        return self._sync(
            self._create_cluster,
            name=name,
            software_environment=software,
            worker_class=worker_class,
            worker_options=worker_options,
            scheduler_options=scheduler_options,
            workspace=workspace or account,
            workers=workers,
            environ=environ,
            tags=tags,
            dask_config=dask_config,
            private_to_creator=private_to_creator,
            scheduler_vm_types=scheduler_vm_types,
            worker_vm_types=worker_vm_types,
            gcp_worker_gpu_type=worker_gpu_type,
            worker_disk_size=worker_disk_size,
            worker_disk_throughput=worker_disk_throughput,
            scheduler_disk_size=scheduler_disk_size,
            backend_options=backend_options,
        )

    @track_context
    async def _delete_cluster(
        self, cluster_id: int, workspace: Optional[str] = None, reason: Optional[str] = None
    ) -> None:
        workspace = workspace or self.default_workspace

        route = f"/api/v2/clusters/account/{workspace}/id/{cluster_id}"
        if reason:
            params = {"reason": reason}
        else:
            params = None
        response = await self._do_request_idempotent(
            "DELETE",
            self.server + route,
            params=params,
        )
        if response.status >= 400:
            await handle_api_exception(response)
        else:
            # multiple deletes sometimes fail if we don't await response here
            await response.json()
            logger.info(f"Cluster {cluster_id} deleted successfully.")

    @overload
    def delete_cluster(
        self: Cloud[Sync],
        cluster_id: int,
        account: Optional[str] = None,
        workspace: Optional[str] = None,
        reason: Optional[str] = None,
    ) -> None: ...

    @overload
    def delete_cluster(
        self: Cloud[Async],
        cluster_id: int,
        account: Optional[str] = None,
        workspace: Optional[str] = None,
        reason: Optional[str] = None,
    ) -> Awaitable[None]: ...

    @delete_docstring  # TODO: this docstring erroneously says "Name of cluster" when it really accepts an ID
    def delete_cluster(
        self,
        cluster_id: int,
        account: Optional[str] = None,
        workspace: Optional[str] = None,
        reason: Optional[str] = None,
    ) -> Optional[Awaitable[None]]:
        return self._sync(self._delete_cluster, cluster_id=cluster_id, workspace=workspace or account, reason=reason)

    async def _get_cluster_state(self, cluster_id: int, workspace: Optional[str] = None) -> dict:
        workspace = workspace or self.default_workspace
        # Make request directly instead of using `_do_request` because we don't want any retries.
        # Retry logic doesn't make sense because this is called by (frequent) period callback, so we'll just wait
        # for next periodic callback call, otherwise retries will overlap with the periodic callback and build up.
        session = self._ensure_session()
        response = await session.request(
            "GET", self.server + f"/api/v2/clusters/account/{workspace}/id/{cluster_id}/state"
        )
        if response.status >= 400:
            await handle_api_exception(response)
        return await response.json()

    async def _get_cluster_details(self, cluster_id: int, workspace: Optional[str] = None):
        workspace = workspace or self.default_workspace
        r = await self._do_request_idempotent(
            "GET", self.server + f"/api/v2/clusters/account/{workspace}/id/{cluster_id}"
        )
        if r.status >= 400:
            await handle_api_exception(r)
        return await r.json()

    def _get_cluster_details_synced(self, cluster_id: int, workspace: Optional[str] = None):
        return self._sync(
            self._get_cluster_details,
            cluster_id=cluster_id,
            workspace=workspace,
        )

    def _cluster_grafana_url(self, cluster_id: int, workspace: Optional[str] = None):
        """for internal Coiled use"""
        workspace = workspace or self.default_workspace
        details = self._sync(
            self._get_cluster_details,
            cluster_id=cluster_id,
            workspace=workspace,
        )

        return get_grafana_url(details, account=workspace, cluster_id=cluster_id)

    def cluster_details(self, cluster_id: int, account: Optional[str] = None, workspace: Optional[str] = None):
        details = self._sync(
            self._get_cluster_details,
            cluster_id=cluster_id,
            workspace=workspace or account,
        )
        state_keys = ["state", "reason", "updated"]

        def get_state(state: dict):
            return {k: v for k, v in state.items() if k in state_keys}

        def get_instance(instance):
            if instance is None:
                return None
            else:
                return {
                    "id": instance["id"],
                    "created": instance["created"],
                    "name": instance["name"],
                    "public_ip_address": instance["public_ip_address"],
                    "private_ip_address": instance["private_ip_address"],
                    "current_state": get_state(instance["current_state"]),
                }

        def get_process(process: dict):
            if process is None:
                return None
            else:
                return {
                    "created": process["created"],
                    "name": process["name"],
                    "current_state": get_state(process["current_state"]),
                    "instance": get_instance(process["instance"]),
                }

        return {
            "id": details["id"],
            "name": details["name"],
            "workers": [get_process(w) for w in details["workers"]],
            "scheduler": get_process(details["scheduler"]),
            "current_state": get_state(details["current_state"]),
            "created": details["created"],
        }

    async def _get_workers_page(self, cluster_id: int, page: int, workspace: Optional[str] = None) -> Tuple[list, bool]:
        page_size = 100
        workspace = workspace or self.default_workspace

        response = await self._do_request(
            "GET",
            self.server + f"/api/v2/workers/account/{workspace}/cluster/{cluster_id}/",
            params={"limit": page_size, "offset": page_size * page},
        )
        if response.status >= 400:
            await handle_api_exception(response)

        results = await response.json()
        has_more_pages = len(results) > 0
        return results, has_more_pages

    @track_context
    async def _get_worker_names(
        self,
        workspace: str,
        cluster_id: int,
        statuses: Optional[List[ProcessStateEnum]] = None,
    ) -> Set[str]:
        worker_infos = await self._depaginate_list(self._get_workers_page, cluster_id=cluster_id, workspace=workspace)
        logger.debug(f"workers: {worker_infos}")
        return {w["name"] for w in worker_infos if statuses is None or w["current_state"]["state"] in statuses}

    @track_context
    async def _security(self, cluster_id: int, workspace: Optional[str] = None, client_wants_public_ip: bool = True):
        cluster_info = await self._get_cluster_details(cluster_id=cluster_id, workspace=workspace)
        if ProcessStateEnum(cluster_info["scheduler"]["current_state"]["state"]) != ProcessStateEnum.started:
            scheduler_state = cluster_info["scheduler"]["current_state"]["state"]
            raise RuntimeError(
                f"Cannot get security info for cluster {cluster_id}, scheduler state is {scheduler_state}"
            )

        public_ip = cluster_info["scheduler"]["instance"]["public_ip_address"]
        private_ip = cluster_info["scheduler"]["instance"]["private_ip_address"]
        tls_cert = cluster_info["cluster_options"]["tls_cert"]
        tls_key = cluster_info["cluster_options"]["tls_key"]
        scheduler_port = cluster_info["scheduler_port"]
        dashboard_address = cluster_info["scheduler"]["dashboard_address"]
        give_scheduler_public_ip = cluster_info["cluster_infra"]["give_scheduler_public_ip"]

        private_address = f"tls://{private_ip}:{scheduler_port}"
        public_address = f"tls://{public_ip}:{scheduler_port}"

        use_public_address = give_scheduler_public_ip and client_wants_public_ip
        if use_public_address:
            if not public_ip:
                raise RuntimeError(
                    "Your Coiled client is configured to use the public IP address, but the scheduler VM does not "
                    "have a public IP address.\n\n"
                    "If you're expecting to connect on private IP address, you can run\n"
                    "    coiled config set coiled.use_scheduler_public_ip False\n"
                    "to configure your local Client to use the private IP address, "
                    "or contact support@coiled.io if you'd like help."
                )
            address_to_use = public_address
        else:
            address_to_use = private_address
            logger.info(f"Connecting to scheduler on its internal address: {address_to_use}")

        # TODO (Declarative): pass extra_conn_args if we care about proxying through Coiled to the scheduler
        security = GatewaySecurity(tls_key, tls_cert)

        return security, {
            "address_to_use": address_to_use,
            "private_address": private_address,
            "public_address": public_address,
            "dashboard_address": dashboard_address,
        }

    @track_context
    async def _requested_workers(self, cluster_id: int, account: Optional[str] = None) -> Set[str]:
        raise NotImplementedError("TODO")

    @overload
    def requested_workers(self: Cloud[Sync], cluster_id: int, account: Optional[str] = None) -> Set[str]: ...

    @track_context
    async def _get_cluster_by_name(self, name: str, workspace: Optional[str] = None) -> int:
        workspace, name = self._normalize_name(
            name, context_workspace=workspace, allow_uppercase=True, property_name="cluster name"
        )

        response = await self._do_request(
            "GET",
            self.server + f"/api/v2/clusters/account/{workspace}/name/{name}",
        )
        if response.status == 404:
            raise DoesNotExist
        elif response.status >= 400:
            await handle_api_exception(response)

        cluster = await response.json()
        return cluster["id"]

    @overload
    def get_cluster_by_name(
        self: Cloud[Sync],
        name: str,
        account: Optional[str] = None,
        workspace: Optional[str] = None,
    ) -> int: ...

    @overload
    def get_cluster_by_name(
        self: Cloud[Async],
        name: str,
        account: Optional[str] = None,
        workspace: Optional[str] = None,
    ) -> Awaitable[int]: ...

    def get_cluster_by_name(
        self,
        name: str,
        account: Optional[str] = None,
        workspace: Optional[str] = None,
    ) -> Union[int, Awaitable[int]]:
        return self._sync(
            self._get_cluster_by_name,
            name=name,
            workspace=workspace or account,
        )

    @track_context
    async def _cluster_status(
        self,
        cluster_id: int,
        account: Optional[str] = None,
        exclude_stopped: bool = True,
    ) -> dict:
        raise NotImplementedError("TODO?")

    @track_context
    async def _get_cluster_states_declarative(
        self,
        cluster_id: int,
        workspace: Optional[str] = None,
        start_time: Optional[datetime.datetime] = None,
    ) -> int:
        workspace = workspace or self.default_workspace

        params = {"start_time": start_time.isoformat()} if start_time is not None else {}

        response = await self._do_request_idempotent(
            "GET",
            self.server + f"/api/v2/clusters/account/{workspace}/id/{cluster_id}/states",
            params=params,
        )
        if response.status >= 400:
            await handle_api_exception(response)

        return await response.json()

    def get_cluster_states(
        self,
        cluster_id: int,
        account: Optional[str] = None,
        workspace: Optional[str] = None,
        start_time: Optional[datetime.datetime] = None,
    ) -> Union[int, Awaitable[int]]:
        return self._sync(
            self._get_cluster_states_declarative,
            cluster_id=cluster_id,
            workspace=workspace or account,
            start_time=start_time,
        )

    def get_clusters_by_name(
        self,
        name: str,
        account: Optional[str] = None,
        workspace: Optional[str] = None,
    ) -> List[dict]:
        """Get all clusters matching name."""
        return self._sync(
            self._get_clusters_by_name,
            name=name,
            workspace=workspace or account,
        )

    @track_context
    async def _get_clusters_by_name(self, name: str, workspace: Optional[str] = None) -> List[dict]:
        workspace, name = self._normalize_name(name, context_workspace=workspace, allow_uppercase=True)

        response = await self._do_request(
            "GET",
            self.server + f"/api/v2/clusters/account/{workspace}/",
            params={"name": name},
        )
        if response.status == 404:
            raise DoesNotExist
        elif response.status >= 400:
            await handle_api_exception(response)

        cluster = await response.json()
        return cluster

    @overload
    def cluster_logs(
        self: Cloud[Sync],
        cluster_id: int,
        account: Optional[str] = None,
        workspace: Optional[str] = None,
        scheduler: bool = True,
        workers: bool = True,
        errors_only: bool = False,
    ) -> Logs: ...

    @overload
    def cluster_logs(
        self: Cloud[Async],
        cluster_id: int,
        account: Optional[str] = None,
        workspace: Optional[str] = None,
        scheduler: bool = True,
        workers: bool = True,
        errors_only: bool = False,
    ) -> Awaitable[Logs]: ...

    @track_context
    async def _cluster_logs(
        self,
        cluster_id: int,
        workspace: Optional[str] = None,
        scheduler: bool = True,
        workers: bool = True,
        errors_only: bool = False,
    ) -> Logs:
        def is_errored(process):
            process_state, instance_state = get_process_instance_state(process)
            return process_state == ProcessStateEnum.error or instance_state == InstanceStateEnum.error

        workspace = workspace or self.default_workspace

        # hits endpoint in order to get scheduler and worker instance names
        cluster_info = await self._get_cluster_details(cluster_id=cluster_id, workspace=workspace)

        try:
            scheduler_name = cluster_info["scheduler"]["instance"]["name"]
        except (TypeError, KeyError):
            # no scheduler instance name in cluster info
            logger.warning("No scheduler found when attempting to retrieve cluster logs.")
            scheduler_name = None

        worker_names = [
            worker["instance"]["name"]
            for worker in cluster_info["workers"]
            if worker["instance"] and (not errors_only or is_errored(worker))
        ]

        LabeledInstance = namedtuple("LabeledInstance", ("name", "label"))

        instances = []
        if scheduler and scheduler_name and (not errors_only or is_errored(cluster_info["scheduler"])):
            instances.append(LabeledInstance(scheduler_name, "Scheduler"))
        if workers and worker_names:
            instances.extend([LabeledInstance(worker_name, worker_name) for worker_name in worker_names])

        async def instance_log_with_semaphor(semaphor, **kwargs):
            async with semaphor:
                return await self._instance_logs(**kwargs)

        # only get 100 logs at a time; the limit here is redundant since aiohttp session already limits concurrent
        # connections but let's be safe just in case
        semaphor = asyncio.Semaphore(value=100)
        results = await asyncio.gather(*[
            instance_log_with_semaphor(semaphor=semaphor, workspace=workspace, instance_name=inst.name)
            for inst in instances
        ])

        out = {
            instance_label: instance_log
            for (_, instance_label), instance_log in zip(instances, results)
            if len(instance_log)
        }

        return Logs(out)

    def cluster_logs(
        self,
        cluster_id: int,
        account: Optional[str] = None,
        workspace: Optional[str] = None,
        scheduler: bool = True,
        workers: bool = True,
        errors_only: bool = False,
    ) -> Union[Logs, Awaitable[Logs]]:
        return self._sync(
            self._cluster_logs,
            cluster_id=cluster_id,
            workspace=workspace or account,
            scheduler=scheduler,
            workers=workers,
            errors_only=errors_only,
        )

    async def _instance_logs(self, workspace: str, instance_name: str, safe=True) -> Log:
        response = await self._do_request(
            "GET",
            self.server + "/api/v2/instances/{}/instance/{}/logs".format(workspace, instance_name),
        )
        if response.status >= 400:
            if safe:
                logger.warning(f"Error retrieving logs for {instance_name}")
                return Log()
            await handle_api_exception(response)

        data = await response.json()

        messages = "\n".join(logline.get("message", "") for logline in data)

        return Log(messages)

    @overload
    def requested_workers(
        self: Cloud[Async], cluster_id: int, account: Optional[str] = None, workspace: Optional[str] = None
    ) -> Awaitable[Set[str]]: ...

    def requested_workers(
        self, cluster_id: int, account: Optional[str] = None, workspace: Optional[str] = None
    ) -> Union[
        Set[str],
        Awaitable[Set[str]],
    ]:
        return self._sync(self._requested_workers, cluster_id, workspace or account)

    @overload
    def scale_up(
        self: Cloud[Sync], cluster_id: int, n: int, account: Optional[str] = None, workspace: Optional[str] = None
    ) -> Optional[Dict]: ...

    @overload
    def scale_up(
        self: Cloud[Async], cluster_id: int, n: int, account: Optional[str] = None, workspace: Optional[str] = None
    ) -> Awaitable[Optional[Dict]]: ...

    def scale_up(
        self, cluster_id: int, n: int, account: Optional[str] = None, workspace: Optional[str] = None
    ) -> Union[Optional[Dict], Awaitable[Optional[Dict]]]:
        """Scale cluster to ``n`` workers

        Parameters
        ----------
        cluster_id
            Unique cluster identifier.
        n
            Number of workers to scale cluster size to.
        account
            **DEPRECATED**. Use ``workspace`` instead.
        workspace
            The Coiled workspace (previously "account") to use. If not specified,
            will check the ``coiled.workspace`` or ``coiled.account`` configuration values,
            or will use your default workspace if those aren't set.

        """
        return self._sync(self._scale_up, cluster_id, n, workspace or account)

    @overload
    def scale_down(
        self: Cloud[Sync],
        cluster_id: int,
        workers: Set[str],
        account: Optional[str] = None,
        workspace: Optional[str] = None,
    ) -> None: ...

    @overload
    def scale_down(
        self: Cloud[Async],
        cluster_id: int,
        workers: Set[str],
        account: Optional[str] = None,
        workspace: Optional[str] = None,
    ) -> Awaitable[None]: ...

    def scale_down(
        self,
        cluster_id: int,
        workers: Set[str],
        account: Optional[str] = None,
        workspace: Optional[str] = None,
    ) -> Optional[Awaitable[None]]:
        """Scale cluster to ``n`` workers

        Parameters
        ----------
        cluster_id
            Unique cluster identifier.
        workers
            Set of workers to scale down to.
        account
            **DEPRECATED**. Use ``workspace`` instead.
        workspace
            The Coiled workspace (previously "account") to use. If not specified,
            will check the ``coiled.workspace`` or ``coiled.account`` configuration values,
            or will use your default workspace if those aren't set.

        """
        return self._sync(self._scale_down, cluster_id, workers, workspace or account)

    @track_context
    async def _better_cluster_logs(
        self,
        cluster_id: int,
        workspace: Optional[str] = None,
        instance_ids: Optional[List[int]] = None,
        dask: bool = False,
        system: bool = False,
        since_ms: Optional[int] = None,
        until_ms: Optional[int] = None,
        filter: Optional[str] = None,
    ):
        workspace = workspace or self.default_workspace

        url_params = {}
        if dask:
            url_params["dask"] = "True"
        if system:
            url_params["system"] = "True"
        if since_ms:
            url_params["since_ms"] = f"{since_ms}"
        if until_ms:
            url_params["until_ms"] = f"{until_ms}"
        if filter:
            url_params["filter_pattern"] = f"{filter}"
        if instance_ids:
            url_params["instance_ids"] = ",".join(map(str, instance_ids))

        url_path = f"/api/v2/clusters/account/{workspace}/id/{cluster_id}/better-logs"

        response = await self._do_request(
            "GET",
            f"{self.server}{url_path}",
            params=url_params,
        )
        if response.status >= 400:
            await handle_api_exception(response)

        data = await response.json()

        return data

    def better_cluster_logs(
        self,
        cluster_id: int,
        account: Optional[str] = None,
        workspace: Optional[str] = None,
        instance_ids: Optional[List[int]] = None,
        dask: bool = False,
        system: bool = False,
        since_ms: Optional[int] = None,
        until_ms: Optional[int] = None,
        filter: Optional[str] = None,
    ) -> Logs:
        return self._sync(
            self._better_cluster_logs,
            cluster_id=cluster_id,
            workspace=workspace or account,
            instance_ids=instance_ids,
            dask=dask,
            system=system,
            since_ms=since_ms,
            until_ms=until_ms,
            filter=filter,
        )

    @track_context
    async def _scale_up(
        self, cluster_id: int, n: int, workspace: Optional[str] = None, reason: Optional[str] = None
    ) -> Dict:
        """
        Increases the number of workers by ``n``.
        """
        workspace = workspace or self.default_workspace
        data = {"n_workers": n}
        if reason:
            # pyright is annoying
            data["reason"] = reason  # type: ignore
        response = await self._do_request(
            "POST", f"{self.server}/api/v2/workers/account/{workspace}/cluster/{cluster_id}/", json=data
        )
        if response.status >= 400:
            await handle_api_exception(response)

        workers_info = await response.json()

        return {"workers": {w["name"] for w in workers_info}}

    @track_context
    async def _scale_down(
        self, cluster_id: int, workers: Iterable[str], workspace: Optional[str] = None, reason: Optional[str] = None
    ) -> None:
        workspace = workspace or self.default_workspace
        workers = list(workers)  # yarl, used by aiohttp, expects list (not set) of strings

        reason_dict = {"reason": reason} if reason else {}
        response = await self._do_request(
            "DELETE",
            f"{self.server}/api/v2/workers/account/{workspace}/cluster/{cluster_id}/",
            params={"name": workers, **reason_dict},
        )
        if response.status >= 400:
            await handle_api_exception(response)

    @overload
    def security(
        self: Cloud[Sync], cluster_id: int, account: Optional[str] = None, workspace: Optional[str] = None
    ) -> Tuple[dask.distributed.Security, dict]: ...

    @overload
    def security(
        self: Cloud[Async], cluster_id: int, account: Optional[str] = None, workspace: Optional[str] = None
    ) -> Awaitable[Tuple[dask.distributed.Security, dict]]: ...

    def security(
        self, cluster_id: int, account: Optional[str] = None, workspace: Optional[str] = None
    ) -> Union[
        Tuple[dask.distributed.Security, dict],
        Awaitable[Tuple[dask.distributed.Security, dict]],
    ]:
        return self._sync(self._security, cluster_id, workspace or account)

    @track_context
    async def _fetch_package_levels(self, workspace: Optional[str] = None) -> List[PackageLevel]:
        workspace = workspace or self.default_workspace
        response = await self._do_request("GET", f"{self.server}/api/v2/packages/", params={"account": workspace})
        if response.status >= 400:
            await handle_api_exception(response)
        return await response.json()

    def get_ssh_key(
        self,
        cluster_id: int,
        workspace: Optional[str] = None,
        worker: Optional[str] = None,
    ) -> dict:
        workspace = workspace or self.default_workspace
        return self._sync(
            self._get_ssh_key,
            cluster_id=cluster_id,
            workspace=workspace,
            worker=worker,
        )

    @track_context
    async def _get_ssh_key(self, cluster_id: int, workspace: str, worker: Optional[str]) -> dict:
        workspace = workspace or self.default_workspace

        route = f"/api/v2/clusters/account/{workspace}/id/{cluster_id}/ssh-key"
        url = f"{self.server}{route}"

        response = await self._do_request("GET", url, params={"worker": worker} if worker else None)
        if response.status >= 400:
            await handle_api_exception(response)
        return await response.json()

    def get_cluster_log_info(
        self,
        cluster_id: int,
        workspace: Optional[str] = None,
    ) -> dict:
        workspace = workspace or self.default_workspace
        return self._sync(
            self._get_cluster_log_info,
            cluster_id=cluster_id,
            workspace=workspace,
        )

    @track_context
    async def _get_cluster_log_info(
        self,
        cluster_id: int,
        workspace: str,
    ) -> dict:
        workspace = workspace or self.default_workspace

        route = f"/api/v2/clusters/account/{workspace}/id/{cluster_id}/log-info"
        url = f"{self.server}{route}"

        response = await self._do_request("GET", url)
        if response.status >= 400:
            await handle_api_exception(response)
        return await response.json()

    @track_context
    async def _get_cluster_aggregated_metric(
        self,
        cluster_id: int,
        workspace: Optional[str],
        query: str,
        over_time: str,
        start_ts: Optional[int],
        end_ts: Optional[int],
    ):
        workspace = workspace or self.default_workspace
        route = f"/api/v2/metrics/account/{workspace}/cluster/{cluster_id}/value"
        url = f"{self.server}{route}"
        params = {"query": query, "over_time": over_time}
        if start_ts:
            params["start_ts"] = str(start_ts)
        if end_ts:
            params["end_ts"] = str(end_ts)

        response = await self._do_request("GET", url, params=params)
        if response.status >= 400:
            await handle_api_exception(response)
        return await response.json()

    @track_context
    async def _add_cluster_span(self, cluster_id: int, workspace: Optional[str], span_identifier: str, data: dict):
        workspace = workspace or self.default_workspace
        route = f"/api/v2/analytics/{workspace}/cluster/{cluster_id}/span/{span_identifier}"
        url = f"{self.server}{route}"

        response = await self._do_request("POST", url, json=data)
        if response.status >= 400:
            await handle_api_exception(response)
        return await response.json()

    def _sync_request(self, route, method: str = "GET", json_result: bool = False, **kwargs):
        url = f"{self.server}{route}"
        response = self._sync(self._do_request, url=url, method=method, **kwargs)
        if response.status >= 400:
            raise ServerError(f"{url} returned {response.status}")

        async def get_result(r):
            return await (r.json() if json_result else r.text())

        return self._sync(
            get_result,
            response,
        )


Cloud = CloudV2


def cluster_logs(
    cluster_id: int,
    account: Optional[str] = None,
    workspace: Optional[str] = None,
    scheduler: bool = True,
    workers: bool = True,
    errors_only: bool = False,
):
    """
    Returns cluster logs as a dictionary, with a key for the scheduler and each worker.

    .. versionchanged:: 0.2.0
       ``cluster_name`` is no longer accepted, use ``cluster_id`` instead.
    """
    with CloudV2() as cloud:
        return cloud.cluster_logs(
            cluster_id=cluster_id,
            workspace=workspace or account,
            scheduler=scheduler,
            workers=workers,
            errors_only=errors_only,
        )


def better_cluster_logs(
    cluster_id: int,
    account: Optional[str] = None,
    workspace: Optional[str] = None,
    instance_ids: Optional[List[int]] = None,
    dask: bool = False,
    system: bool = False,
    since_ms: Optional[int] = None,
    until_ms: Optional[int] = None,
    filter: Optional[str] = None,
):
    """
    Pull logs for the cluster using better endpoint.

    Logs for recent clusters are split between system and container (dask), you can get
    either or both (or none).

    since_ms and until_ms are both inclusive (you'll get logs with timestamp matching those).
    """
    with Cloud() as cloud:
        return cloud.better_cluster_logs(
            cluster_id=cluster_id,
            workspace=workspace or account,
            instance_ids=instance_ids,
            dask=dask,
            system=system,
            since_ms=since_ms,
            until_ms=until_ms,
            filter=filter,
        )


def cluster_details(
    cluster_id: int,
    account: Optional[str] = None,
    workspace: Optional[str] = None,
) -> dict:
    """
    Get details of a cluster as a dictionary.
    """
    with CloudV2() as cloud:
        return cloud.cluster_details(
            cluster_id=cluster_id,
            workspace=workspace or account,
        )


def log_cluster_debug_info(
    cluster_id: int,
    account: Optional[str] = None,
    workspace: Optional[str] = None,
):
    with CloudV2() as cloud:
        details = cloud.cluster_details(cluster_id, workspace or account)
        logger.debug("Cluster details:")
        logger.debug(json.dumps(details, indent=2))

        states_by_type = cloud.get_cluster_states(cluster_id, workspace or account)

        logger.debug("cluster state history:")
        log_states(flatten_log_states(states_by_type), level=logging.DEBUG)


def create_cluster(
    name: str,
    *,
    software: Optional[str] = None,
    worker_options: Optional[dict] = None,
    scheduler_options: Optional[dict] = None,
    account: Optional[str] = None,
    workspace: Optional[str] = None,
    workers: int = 0,
    environ: Optional[Dict] = None,
    tags: Optional[Dict] = None,
    dask_config: Optional[Dict] = None,
    private_to_creator: Optional[bool] = None,
    scheduler_vm_types: Optional[list] = None,
    worker_vm_types: Optional[list] = None,
    worker_disk_size: Optional[int] = None,
    worker_disk_throughput: Optional[int] = None,
    scheduler_disk_size: Optional[int] = None,
    backend_options: Optional[Union[dict, AWSOptions, GCPOptions]] = None,
) -> int:
    """Create a cluster

    Parameters
    ---------
    name
        Name of cluster.
    software
        Identifier of the software environment to use, in the format (<account>/)<name>. If the software environment
        is owned by the same account as that passed into "account", the (<account>/) prefix is optional.

        For example, suppose your account is "wondercorp", but your friends at "friendlycorp" have an environment
        named "xgboost" that you want to use; you can specify this with "friendlycorp/xgboost". If you simply
        entered "xgboost", this is shorthand for "wondercorp/xgboost".

        The "name" portion of (<account>/)<name> can only contain ASCII letters, hyphens and underscores.
    worker_options
        Mapping with keyword arguments to pass to ``worker_class``. Defaults to ``{}``.
    scheduler_options
        Mapping with keyword arguments to pass to the Scheduler ``__init__``. Defaults to ``{}``.
    account
        **DEPRECATED**. Use ``workspace`` instead.
    workspace
        The Coiled workspace (previously "account") to use. If not specified,
        will check the ``coiled.workspace`` or ``coiled.account`` configuration values,
        or will use your default workspace if those aren't set.
    workers
        Number of workers we to launch.
    environ
        Dictionary of environment variables.
    tags
        Dictionary of instance tags
    dask_config
        Dictionary of dask config to put on cluster

    See Also
    --------
    coiled.Cluster
    """
    with CloudV2(account=workspace or account) as cloud:
        cluster, existing = cloud.create_cluster(
            name=name,
            software=software,
            worker_options=worker_options,
            scheduler_options=scheduler_options,
            workspace=workspace or account,
            workers=workers,
            environ=environ,
            tags=tags,
            dask_config=dask_config,
            private_to_creator=private_to_creator,
            backend_options=backend_options,
            worker_vm_types=worker_vm_types,
            worker_disk_size=worker_disk_size,
            worker_disk_throughput=worker_disk_throughput,
            scheduler_disk_size=scheduler_disk_size,
            scheduler_vm_types=scheduler_vm_types,
        )
        return cluster


@list_docstring
def list_clusters(account=None, workspace=None, max_pages: Optional[int] = None):
    with CloudV2() as cloud:
        return cloud.list_clusters(workspace=workspace or account, max_pages=max_pages)


@delete_docstring
def delete_cluster(name: str, account: Optional[str] = None, workspace: Optional[str] = None):
    with CloudV2() as cloud:
        cluster_id = cloud.get_cluster_by_name(name=name, workspace=workspace or account)
        if cluster_id is not None:
            return cloud.delete_cluster(cluster_id=cluster_id, workspace=workspace or account)



import datetime
import os
import re
import shutil
import subprocess
from pathlib import Path
from typing import Optional

import click
import dask.config
from rich import print
from rich.prompt import Confirm

import coiled
from coiled.utils import get_temp_dir

from ..utils import CONTEXT_SETTINGS
from .utils import find_cluster

DEFAULT_SSH_AGENT_CONFIG = "Host *.dask.host\n  IdentityAgent SSH_AUTH_SOCK"


def open_ssh(address: str, key: str, jump_to_address: Optional[str] = None, host_command: Optional[str] = None):
    """Open an SSH session, relies on `ssh` and `ssh-add` (agent)."""
    if not add_key_to_agent(address, key, t=5):
        return

    ssh_shell(address, jump_to_address, host_command)


def get_ssh_path() -> Optional[str]:
    ssh_path = shutil.which("ssh")
    if not ssh_path:
        print("Unable to find `ssh`, you may need to install OpenSSH or add it to your paths.")
        return None
    return ssh_path


def get_ssh_add_path() -> Optional[str]:
    ssh_add_path = shutil.which("ssh-add")
    if not ssh_add_path:
        print("Unable to find `ssh-add`, you may need to install OpenSSH or add it to your paths.")
        return None
    return ssh_add_path


def check_ssh_agent():
    user_config_path = Path(os.path.expanduser("~")) / ".ssh" / "config"
    if user_config_path.exists():
        with open(user_config_path) as f:
            user_config = f.read()
            if "dask.host" in user_config:
                return True
            if re.search("IdentityAgent[^\n]+1password", user_config):
                print("Coiled cannot add temporary SSH key to [bold]1Password SSH Agent[/bold].")
                print(
                    "You can explicitly allow Coiled to use the system SSH agent by adding these lines to the "
                    f"beginning of [green]{user_config_path}[/green]:\n\n{DEFAULT_SSH_AGENT_CONFIG}\n\n"
                )
                if Confirm.ask("May Coiled edit your ssh config to add these lines?", default=True):
                    with open(user_config_path, "w") as f:
                        f.write(f"{DEFAULT_SSH_AGENT_CONFIG}\n\n{user_config}")
                    return True
                return False
    return True


def add_key_to_agent(address: str, key: str, t: Optional[int] = None, delete: bool = False) -> bool:
    """Add (or remove) ssh key to ssh agent."""
    ssh_add_path = get_ssh_add_path()
    if not ssh_add_path:
        return False

    with get_temp_dir() as keydir:
        key_path = os.path.join(keydir, f"scheduler-key-{address}.pem")

        with open(key_path, "w") as f:
            f.write(key)

        # ssh needs file permissions to be set
        os.chmod(key_path, mode=0o600)

        cmd = [ssh_add_path]
        if delete:
            cmd.append("-d")
        if t:
            cmd.extend(["-t", f"{t}"])
        cmd.append(key_path)

        # briefly add key to agent, this allows us to jump to worker with agent forwarding
        p = subprocess.run(cmd, capture_output=True, text=True)

        if p.returncode and p.stderr:
            print("An error occurred calling `ssh-add`. You may need to enable the ssh agent.")
            print(p.stderr)
            return False

    return True


def _write_coiled_ssh_key(coiled_ssh_dir, address, key):
    key_path = os.path.join(coiled_ssh_dir, f"scheduler-key-{address}.pem")

    with open(key_path, "w") as f:
        f.write(key)
    # ssh needs file permissions to be set
    os.chmod(key_path, mode=0o600)

    return key_path


def _delete_coiled_ssh_key_file(coiled_ssh_dir, address):
    key_path = os.path.join(coiled_ssh_dir, f"scheduler-key-{address}.pem")

    if os.path.exists(key_path):
        os.remove(key_path)


def _add_host_to_coiled_ssh_config(coiled_config_path, address, key_path):
    identity_config = f"""
Host {address}
    # added {datetime.datetime.now()}
    HostName {address}
    User ubuntu
    IdentityFile {key_path}
    """
    # print(identity_config)
    do_write = False
    if not os.path.exists(coiled_config_path):
        do_write = True
    else:
        with open(coiled_config_path, "r") as f:
            if f"Host {address}" not in f.read():
                do_write = True
            else:
                # print(f"Host {address} is already in {coiled_config_path}")
                ...
    if do_write:
        with open(coiled_config_path, "a") as af:
            af.write(identity_config)


def _remove_host_from_coiled_ssh_config(coiled_config_path, address):
    if os.path.exists(coiled_config_path):
        with open(coiled_config_path, "r") as f:
            contents = f.read()
        if f"Host {address}" in contents:
            updated = re.sub(rf"\nHost {address}(\n +[^\n]+)*\n", "", contents, flags=re.MULTILINE)
            with open(coiled_config_path, "w") as f:
                f.write(updated)
            # print(f"Removed 'Host {address} [...]' from {coiled_config_path}")


def _add_coiled_to_ssh_config(coiled_config_path):
    ssh_dir = Path(os.path.expanduser("~")) / ".ssh"
    ssh_config_path = ssh_dir / "config"
    include_line = f"Include {coiled_config_path}"
    try:
        os.makedirs(ssh_dir, exist_ok=True)
        if ssh_config_path.exists():
            with open(ssh_config_path, "r") as f:
                ssh_config_contents = f.read()
        else:
            ssh_config_contents = ""

        if include_line not in ssh_config_contents:
            with open(ssh_config_path, "w") as wf:
                print(f"Add {include_line!r} to {ssh_config_path}")
                wf.write(f"{include_line}\n\n{ssh_config_contents}")
        else:
            # print(f"{include_line!r} already in {ssh_config_path}")
            ...
    except Exception as e:
        raise coiled.exceptions.CoiledException(
            "Unable to update your SSH config file to include the Coiled-specific SSH config because of this error:\n"
            f"  {e}\n\n"
            "You can run\n"
            "  coiled config set coiled.use_ssh_agent True\n"
            "to use the SSH agent instead of SSH config for Coiled SSH access,\n"
            "or contact support@coiled.io if you'd like help troubleshooting."
        ) from None


def add_ssh_identity_to_config(address: str, key: str):
    # Make directory for Coiled-specific SSH things (keys and config)
    coiled_ssh_dir = Path(os.path.expanduser("~")) / ".coiled" / "ssh"
    os.makedirs(coiled_ssh_dir, exist_ok=True)

    # Key
    key_path = _write_coiled_ssh_key(coiled_ssh_dir, address, key)

    # Add host to the Coiled-specific SSH config file
    coiled_config_path = os.path.join(coiled_ssh_dir, "coiled-identities")
    _add_host_to_coiled_ssh_config(coiled_config_path, address, key_path)

    # Make sure Coiled-specific SSH config file is included in general SSH config file
    _add_coiled_to_ssh_config(coiled_config_path)


def remove_ssh_key_from_config(address: str):
    # Remove the SSH key file
    coiled_ssh_dir = Path(os.path.expanduser("~")) / ".coiled" / "ssh"
    _delete_coiled_ssh_key_file(coiled_ssh_dir, address)

    # Remove identity from Coiled-specific SSH config file
    coiled_config_path = os.path.join(coiled_ssh_dir, "coiled-identities")
    _remove_host_from_coiled_ssh_config(coiled_config_path, address)


def add_ssh_access(address: str, key: str, t=None):
    if dask.config.get("coiled.use_ssh_agent", False):
        add_key_to_agent(address, key, t)
    else:
        add_ssh_identity_to_config(address, key)


def remove_ssh_access(address: str, key: str):
    if dask.config.get("coiled.use_ssh_agent", False):
        add_key_to_agent(address, key, delete=True)
    else:
        remove_ssh_key_from_config(address)


def ssh_shell(address: str, jump_to_address: Optional[str] = None, host_command: Optional[str] = None):
    if jump_to_address:
        ssh_target = ["-J", f"ubuntu@{address}", f"ubuntu@{jump_to_address}"]
        ssh_target_label = f"worker at {jump_to_address}"
    else:
        ssh_target = [f"ubuntu@{address}"]
        ssh_target_label = f"scheduler at {address}"

    ssh_path = get_ssh_path()
    if not ssh_path:
        return
    ssh_command = [ssh_path, "-t", *ssh_target, "-o", "StrictHostKeyChecking=no", "-o", "ForwardAgent=yes"]
    if host_command:
        ssh_command.append(host_command)

    print(f"===Starting SSH session to {ssh_target_label}===")
    if host_command:
        print(f"> {host_command}")
    subprocess.run(ssh_command)
    print("===SSH session closed===")


@click.command(context_settings=CONTEXT_SETTINGS)
@click.argument("cluster", default="", required=False)
@click.option(
    "--account",
    "--workspace",
    default=None,
    help="Coiled workspace (uses default workspace if not specified)."
    " Note: --account is deprecated, please use --workspace instead.",
)
@click.option(
    "--private",
    default=False,
    is_flag=True,
    help="Use private IP address of scheduler (default is DNS hostname for public IP)",
)
@click.option(
    "--by-ip",
    default=False,
    is_flag=True,
    help="Use public IP address of scheduler directly, not using DNS hostname",
)
@click.option(
    "--worker",
    default=None,
    help="Connect to worker with specified name or private IP address (default is to connect to scheduler)",
)
@click.option(
    "--add-key",
    default=False,
    is_flag=True,
    help="Just add ssh key to local OpenSSH agent, no lifetime/expiration set",
)
@click.option(
    "--delete-key",
    default=False,
    is_flag=True,
    help="Just delete ssh key from local OpenSSH agent",
)
@click.option(
    "--dask", default=False, is_flag=True, help="Attach to shell in Dask container rather than shell on host machine."
)
@click.option(
    "--command",
    default=None,
    hidden=True,
)
def ssh(
    cluster: str,
    account: Optional[str],
    private: bool,
    by_ip: bool,
    worker: Optional[str],
    add_key: bool,
    delete_key: bool,
    dask: bool,
    command: Optional[str],
):
    with coiled.Cloud(account=account) as cloud:
        cluster_info = find_cluster(cloud, cluster)
        cluster_id = cluster_info["id"]
        ssh_info = cloud.get_ssh_key(cluster_id=cluster_id, worker=worker)

    if private:
        scheduler_address = ssh_info["scheduler_private_address"]
    else:
        if by_ip:
            scheduler_address = ssh_info["scheduler_public_address"]
        else:
            scheduler_address = ssh_info["scheduler_hostname"] or ssh_info["scheduler_public_address"]

    if add_key:
        # just add the key, maybe you want to use rsync or something like that
        add_ssh_access(scheduler_address, key=ssh_info["private_key"])

    elif delete_key:
        # just delete the key
        remove_ssh_access(scheduler_address, key=ssh_info["private_key"])

    else:
        if not check_ssh_agent():
            print("You may get an error, but we'll try SSH anyway...")

        # open an ssh shell
        if ssh_info["scheduler_state"] not in ("starting", "started"):
            print(
                f"[red]Scheduler for {cluster_info['name']} ({cluster_id}) is "
                f"[bold]{ssh_info['scheduler_state']}[/bold]. "
                "You can only connect when scheduler is 'starting' or 'started'."
            )
            return

        if not scheduler_address:
            print(
                f"[red]Scheduler for {cluster_info['name']} ({cluster_id}) is "
                f"[bold]{ssh_info['scheduler_state']}[/bold] and does not currently have an IP address.[/]"
            )
            return

        if not ssh_info["private_key"]:
            print("Unable to retrieve SSH key")
            return

        # Default container name is parent dir (tmp) + service name (dask) + instance number (1).
        # If we need to make this more flexible in the future, name could be returned by cloud.get_ssh_key;
        # for now we don't need that flexibility and it's okay if SSH CLI requires client update someday.
        command = f"docker exec -it tmp-dask-1 {command or 'bash'}" if dask else command

        open_ssh(
            address=scheduler_address,
            key=ssh_info["private_key"],
            jump_to_address=ssh_info["worker_address"],
            host_command=command,
        )



import asyncio
import os
import platform
import re
import shutil
import subprocess
from collections import defaultdict
from concurrent.futures import ThreadPoolExecutor
from contextlib import asynccontextmanager
from logging import getLogger
from pathlib import Path
from sys import executable
from threading import Lock
from typing import Dict, List, Optional, TextIO, Tuple, Union, cast
from urllib.parse import urlparse

import toml
from filelock import BaseFileLock, FileLock
from packaging.utils import parse_wheel_filename
from rich.progress import Progress
from typing_extensions import Literal
from urllib3.util import parse_url

from coiled.context import track_context
from coiled.types import PackageInfo, PackageLevelEnum, ResolvedPackageInfo
from coiled.utils import (
    TemporaryDirectory,
    get_encoding,
    get_temp_dir,
    partition,
    recurse_importable_python_files,
    validate_wheel,
)
from coiled.v2.widgets.util import simple_progress

try:
    import keyring  # type: ignore[reportMissingImports]

    HAVE_KEYRING = True
except Exception:
    HAVE_KEYRING = False


logger = getLogger("coiled.software_utils")
subdir_datas = {}

ANY_AVAILABLE = "ANY-AVAILABLE"
COILED_LOCAL_PACKAGE_PREFIX = "coiled_local_"
DEFAULT_PYPI_URL = "https://pypi.org/simple"
DEFAULT_JSON_PYPI_URL = "https://pypi.org/pypi"
PYTHON_VERSION = platform.python_version_tuple()
LOCK_PATH = Path(os.path.expanduser("~")) / ".coiled" / "locks"


async def create_subprocess_exec(
    program: str,
    *args: str,
    stdout: Union[TextIO, int, None] = None,
    stderr: Union[TextIO, int, None] = None,
    extra_env: Optional[Dict[str, str]] = None,
) -> subprocess.CompletedProcess:
    # create_subprocess_exec is broken with IPython on Windows,
    # because it uses the wrong event loop
    loop = asyncio.get_running_loop()
    env = {**os.environ, **(extra_env or {})}
    result = loop.run_in_executor(
        None, lambda: subprocess.run([program, *args], stdout=stdout, stderr=stderr, close_fds=True, env=env)
    )
    return await result


def partition_ignored_packages(
    packages: List[PackageInfo], priorities: Dict[Tuple[str, Literal["conda", "pip"]], PackageLevelEnum]
) -> Tuple[List[PackageInfo], List[PackageInfo]]:
    return partition(
        packages,
        lambda pkg: priorities.get((pkg["name"], pkg["source"])) == PackageLevelEnum.IGNORE,
    )


def partition_local_python_code_packages(packages: List[PackageInfo]) -> Tuple[List[PackageInfo], List[PackageInfo]]:
    return partition(
        packages,
        lambda pkg: (
            pkg["name"].startswith(COILED_LOCAL_PACKAGE_PREFIX) and not cast(str, pkg["wheel_target"]).endswith(".whl")
        ),
    )


def partition_local_packages(packages: List[PackageInfo]) -> Tuple[List[PackageInfo], List[PackageInfo]]:
    return partition(
        packages,
        lambda pkg: bool(pkg["wheel_target"]),
    )


WHEEL_BUILD_LOCKS: Dict[str, Tuple[BaseFileLock, Lock, TemporaryDirectory]] = {}


# filelock is thread local
# so we have to ensure the lock is acquired/released
# on the same thread
FILE_LOCK_POOL = ThreadPoolExecutor(max_workers=1)
THREAD_LOCK_POOL = ThreadPoolExecutor(max_workers=1)


@asynccontextmanager
async def async_lock(file_lock: BaseFileLock, thread_lock: Lock):
    # Beware, there are some complicated details to this locking implementation!
    # We're trying to manage the weirdness of the file lock mostly.
    loop = asyncio.get_event_loop()
    # first acquire a thread lock
    await loop.run_in_executor(THREAD_LOCK_POOL, thread_lock.acquire)
    # acquire the file lock, we should be the only thread trying to get it
    # the threadpool is required to release it, so another thread
    # attempting to get the lock will deadlock things by preventing the
    # release!
    await loop.run_in_executor(FILE_LOCK_POOL, file_lock.acquire)
    yield
    # release the file lock first
    await loop.run_in_executor(FILE_LOCK_POOL, file_lock.release)
    # now release the thread lock, allowing another thread to proceed
    # and get the file lock
    thread_lock.release()


@track_context
async def create_wheel(pkg_name: str, version: str, src: str) -> ResolvedPackageInfo:
    # These locks are set up such that
    # Threads: Block on each other and check if another thread already built the wheel
    # Processes: Block on each other, but will not reuse a wheel created by another
    # `pip wheel` is never run on the same package at the same time
    LOCK_PATH.mkdir(parents=True, exist_ok=True)  # ensure lockfile directory exists
    package_lock, thread_lock, tmpdir = WHEEL_BUILD_LOCKS.setdefault(
        pkg_name,
        (
            FileLock(LOCK_PATH / ("." + pkg_name + version + ".build-lock")),
            Lock(),
            get_temp_dir(ignore_cleanup_errors=True),
        ),
    )
    async with async_lock(package_lock, thread_lock):
        outdir = Path(tmpdir.name) / Path(pkg_name)
        if outdir.exists():
            logger.debug(f"Checking for existing wheel for {pkg_name} @ {outdir}")
            wheel_fn = next((file for file in outdir.iterdir() if file.suffix == ".whl"), None)
        else:
            wheel_fn = None
        if not wheel_fn:
            logger.debug(f"No existing wheel, creating a wheel for {pkg_name} @ {src}")
            # must use executable to avoid using some other random python
            proc = await create_subprocess_exec(
                executable,
                "-m",
                "pip",
                "wheel",
                "--wheel-dir",
                str(outdir),
                "--no-deps",
                "--use-pep517",
                "--no-cache-dir",
                src,
                stderr=subprocess.STDOUT,
                stdout=subprocess.PIPE,
                extra_env={"PIP_REQUIRE_VIRTUALENV": "false"},
            )
            if proc.returncode:
                print(f"---Wheel Build Log for {pkg_name}---\n" + proc.stdout.decode(encoding=get_encoding()))
                return {
                    "name": pkg_name,
                    "source": "pip",
                    "channel": None,
                    "conda_name": None,
                    "client_version": version,
                    "specifier": "",
                    "include": False,
                    "error": (
                        "Failed to build a wheel for the"
                        " package, will not be included in environment, check stdout for the build log"
                    ),
                    "note": None,
                    "sdist": None,
                    "md5": None,
                }
            wheel_fn = next(file for file in outdir.iterdir() if file.suffix == ".whl")
        logger.debug(f"Using wheel @ {wheel_fn}")
        _, build_version, _, _ = parse_wheel_filename(str(wheel_fn.name))
        has_python, md5, missing_py_files = await validate_wheel(wheel_fn, src)
    return {
        "name": pkg_name,
        "source": "pip",
        "channel": None,
        "conda_name": None,
        "client_version": str(build_version),
        "specifier": "",
        "include": True,
        "error": None if has_python else "Built wheel contains no python files!",
        "note": (
            f"Wheel built from {src}"
            + (f" is missing {', '.join(sorted(missing_py_files)[:10])}" if missing_py_files else "")
        ),
        "sdist": wheel_fn.open("rb"),
        "md5": md5,
    }


@track_context
async def create_wheel_from_egg(pkg_name: str, version: str, src: str) -> ResolvedPackageInfo:
    tmpdir = get_temp_dir(ignore_cleanup_errors=True)
    outdir = Path(tmpdir.name) / Path(pkg_name)
    outdir.mkdir(parents=True)
    logger.debug(f"Attempting to create a wheel for {pkg_name} in directory {src}")
    # must use executable to avoid using some other random python
    proc = await create_subprocess_exec(
        executable,
        "-m",
        "wheel",
        "convert",
        "--dest-dir",
        str(outdir),
        src,
        stderr=subprocess.STDOUT,
        stdout=subprocess.PIPE,
        extra_env={"PIP_REQUIRE_VIRTUALENV": "false"},
    )
    if proc.returncode:
        print(f"---Egg to wheel conversion Log for {pkg_name}---\n" + proc.stdout.decode(encoding=get_encoding()))
        return {
            "name": pkg_name,
            "source": "pip",
            "channel": None,
            "conda_name": None,
            "client_version": version,
            "specifier": "",
            "include": False,
            "error": (
                "Failed to convert the package egg to a wheel"
                ", will not be included in environment, check stdout for egg conversion log"
            ),
            "note": None,
            "sdist": None,
            "md5": None,
        }
    wheel_fn = next(file for file in outdir.iterdir() if file.suffix == ".whl")
    has_python, md5, missing_py_files = await validate_wheel(Path(wheel_fn), tmpdir.name)
    return {
        "name": pkg_name,
        "source": "pip",
        "channel": None,
        "conda_name": None,
        "client_version": version,
        "specifier": "",
        "include": True,
        "error": None if has_python else "Built wheel has no python files!",
        "note": (
            "Wheel built from local egg"
            + (f" is missing {', '.join(sorted(missing_py_files)[:10])}" if missing_py_files else "")
        ),
        "sdist": wheel_fn.open("rb"),
        "md5": md5,
    }


@track_context
async def create_wheel_from_src_dir(pkg_name: str, version: str, src: str) -> ResolvedPackageInfo:
    # These locks are set up such that
    # Threads: Block on each other and check if another thread already built the tarball
    # Processes: Block on each other, but will not reuse a tarball created by another
    md5 = None
    LOCK_PATH.mkdir(parents=True, exist_ok=True)  # ensure lockfile directory exists
    package_lock, thread_lock, tmpdir = WHEEL_BUILD_LOCKS.setdefault(
        pkg_name,
        (
            FileLock(LOCK_PATH / (f".{pkg_name}{version}.build-lock")),
            Lock(),
            get_temp_dir(ignore_cleanup_errors=True),
        ),
    )
    async with async_lock(package_lock, thread_lock):
        outdir = Path(tmpdir.name) / Path(pkg_name)
        if outdir.exists():
            logger.debug(f"Checking for existing source archive for {pkg_name} @ {outdir}")
            wheel_fn = next((file for file in outdir.iterdir() if file.suffix == ".whl"), None)
        else:
            wheel_fn = None
        if not wheel_fn:
            logger.debug(f"No existing source archive, creating an archive for {pkg_name} @ {src}")
            try:
                unpacked_dir = outdir / f"{pkg_name}-{version}"
                # Create fake metadata to make wheel work
                dist_info_dir = unpacked_dir / f"{unpacked_dir.name}.dist-info"
                dist_info_dir.mkdir(parents=True)
                with open(dist_info_dir / "METADATA", "w") as f:
                    f.write(f"Metadata-Version: 2.1\nName: {pkg_name}\nVersion: {version}\n")
                with open(dist_info_dir / "WHEEL", "w") as f:
                    f.write("Wheel-Version: 1.0\nGenerator: coiled\nRoot-Is-Purelib: true\nTag: py3-none-any\n")
                src_path = Path(src)
                for file in recurse_importable_python_files(src_path):
                    if str(file) in ("__init__.py", "__main__.py"):
                        continue
                    dest = unpacked_dir / file
                    dest.parent.mkdir(parents=True, exist_ok=True)
                    shutil.copy(src_path / file, dest)
                proc = await create_subprocess_exec(
                    executable,
                    "-m",
                    "wheel",
                    "pack",
                    "--dest-dir",
                    str(outdir),
                    str(unpacked_dir),
                    stderr=subprocess.STDOUT,
                    stdout=subprocess.PIPE,
                    extra_env={"PIP_REQUIRE_VIRTUALENV": "false"},
                )
                if proc.returncode:
                    print(f"---wheel packing log for {src}---\n" + proc.stdout.decode(encoding=get_encoding()))
                    return {
                        "name": pkg_name,
                        "source": "pip",
                        "channel": None,
                        "conda_name": None,
                        "client_version": version,
                        "specifier": "",
                        "include": False,
                        "error": (
                            "Failed to build a package of your local python files. Please check stdout for details"
                        ),
                        "note": None,
                        "sdist": None,
                        "md5": None,
                    }
            except IOError as e:
                return {
                    "name": pkg_name,
                    "source": "pip",
                    "channel": None,
                    "conda_name": None,
                    "client_version": version,
                    "specifier": "",
                    "include": False,
                    "error": f"Failed to build a package of your local python files. Exception: {e}",
                    "note": None,
                    "sdist": None,
                    "md5": None,
                }
            wheel_fn = next(file for file in outdir.iterdir() if file.suffix == ".whl")
        logger.debug(f"Using wheel @ {wheel_fn}")
        _, build_version, _, _ = parse_wheel_filename(str(wheel_fn.name))
        has_python, md5, missing_py_files = await validate_wheel(wheel_fn, src)
    return {
        "name": pkg_name,
        "source": "pip",
        "channel": None,
        "conda_name": None,
        "client_version": str(build_version),
        "specifier": "",
        "include": True,
        "error": None if has_python else "Built wheel does not contain all python files!",
        "note": (
            f"Source wheel built from {src}"
            + (f" is missing {', '.join(sorted(missing_py_files)[:10])}" if missing_py_files else "")
        ),
        "sdist": wheel_fn.open("rb"),
        "md5": md5,
    }


async def create_wheels_for_local_python(packages: List[PackageInfo], progress: Optional[Progress] = None):
    finalized_packages: list[ResolvedPackageInfo] = []
    home_dir = str(Path.home())
    for pkg in packages:
        if pkg["wheel_target"]:
            with simple_progress(
                f'Creating wheel for {pkg["wheel_target"].replace(home_dir, "~", 1)}', progress=progress
            ):
                finalized_packages.append(
                    await create_wheel_from_src_dir(
                        pkg_name=pkg["name"],
                        version=pkg["version"],
                        src=pkg["wheel_target"],
                    )
                )
    return finalized_packages


async def create_wheels_for_packages(
    packages: List[PackageInfo],
    progress: Optional[Progress] = None,
):
    finalized_packages: list[ResolvedPackageInfo] = []
    for pkg in packages:
        if pkg["wheel_target"]:
            if pkg["wheel_target"].endswith(".egg"):
                with simple_progress(f'Creating wheel from egg for {pkg["name"]}', progress=progress):
                    finalized_packages.append(
                        await create_wheel_from_egg(
                            pkg_name=pkg["name"],
                            version=pkg["version"],
                            src=pkg["wheel_target"],
                        )
                    )
            else:
                with simple_progress(f'Creating wheel for {pkg["name"]}', progress=progress):
                    finalized_packages.append(
                        await create_wheel(
                            pkg_name=pkg["name"],
                            version=pkg["version"],
                            src=pkg["wheel_target"],
                        )
                    )
    return finalized_packages


pip_bad_req_regex = (
    r"(?P<package>.+) (?P<version>.+) has requirement "
    r"(?P<requirement>.+), but you have (?P<requirement2>.+) (?P<reqversion>.+)\."
)


@track_context
async def check_pip_happy(progress: Optional[Progress] = None) -> Dict[str, List[str]]:
    with simple_progress("Running pip check", progress=progress):
        proc = await create_subprocess_exec(
            executable,
            "-m",
            "pip",
            "check",
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            extra_env={"PIP_REQUIRE_VIRTUALENV": "false"},
        )
        faulty_packages = defaultdict(list)
        if proc.returncode:
            output = proc.stdout.decode(encoding=get_encoding())
            bad_reqs = re.finditer(pip_bad_req_regex, output)
            for bad_req in bad_reqs:
                groups = bad_req.groupdict()
                span = bad_req.span()
                warning = output[span[0] : span[1]]
                faulty_packages[groups["package"]].append(warning)
        return faulty_packages


def _load_toml(path: Union[Path, str]) -> Dict:
    path = Path(path)
    filename = path.name
    dir_path = path.parent
    if dir_path.exists():
        dir_path = dir_path.resolve()
    toml_dict = {}
    # Walk up the directory tree to find the first toml file (up to 10 levels above)
    for _ in range(10):
        path = dir_path / filename
        if not path.exists():
            if dir_path == dir_path.parent:
                break
            dir_path = dir_path.parent
            continue
        try:
            toml_dict = toml.load(path.open())
        except toml.TomlDecodeError as e:
            logger.debug(f"Failed to load {path}: {e}", exc_info=True)
        break
    return toml_dict


def _get_pip_index_urls() -> List[str]:
    """Returns the index URLs from `pip config list` as a list of strings."""
    encoding = get_encoding()

    index_url = DEFAULT_PYPI_URL
    extra_index_urls = []

    # Check pip config first
    try:
        config_output = subprocess.check_output(
            [
                executable,
                "-m",
                "pip",
                "--no-input",
                "config",
                "list",
            ],
            encoding=encoding,
            env={**os.environ, "PIP_REQUIRE_VIRTUALENV": "false"},
        )
        for line in config_output.splitlines():
            if ".index-url" in line:
                index_url = line.split("=", 1)[1].strip("' \n\"")
            if ".extra-index-url" in line:
                extra_index_urls.append(line.split("=", 1)[1].strip("' \n\""))

    except (subprocess.CalledProcessError, FileNotFoundError):
        pass

    # Load pyproject.toml to check other tool configs
    pyproject = _load_toml("pyproject.toml")
    tool_dict = pyproject.get("tool", {})

    # poetry
    poetry_sources = tool_dict.get("poetry", {}).get("source", [])
    for source in poetry_sources:
        if source_url := source.get("url"):
            extra_index_urls.append(source_url)
    # uv
    uv_conf = {**tool_dict.get("uv", {}), **_load_toml("uv.toml")}.get("pip", {})
    index_url = uv_conf.get("index-url") or index_url
    extra_index_urls.extend(uv_conf.get("extra-index-url", "").split())

    # pixi
    pixi_conf = {**tool_dict.get("pixi", {}), **_load_toml("pixi.toml")}.get("pypi-options", {})
    index_url = pixi_conf.get("index-url") or index_url
    extra_index_urls.extend(pixi_conf.get("extra-index-urls", []))

    # Environment variables take precedence over config files
    # See https://pip.pypa.io/en/stable/topics/configuration/#precedence-override-order
    extra_index_urls.extend(os.environ.get("PIP_EXTRA_INDEX_URL", "").split())
    extra_index_urls.extend(os.environ.get("UV_EXTRA_INDEX_URL", "").split())
    index_url = os.environ.get("PIP_INDEX_URL", index_url)
    index_url = os.environ.get("UV_INDEX_URL", index_url)

    extra_index_urls = sorted(set(extra_index_urls) - {index_url})

    return [url.strip() for url in [index_url, *extra_index_urls]]


def get_index_urls():
    index_urls = [(DEFAULT_PYPI_URL if url == DEFAULT_JSON_PYPI_URL else url) for url in _get_pip_index_urls()]

    # Include netrc auth in URLs if available
    authed_index_urls = []
    for index_url in index_urls:
        # Do not bother checking for passwords to pypi.org
        if index_url != DEFAULT_PYPI_URL:
            try:
                parsed_url = parse_url(index_url)
            except Exception:
                logger.warning(f"Failed to parse PyPI index URL {index_url}. Skipping URL")
                continue
            username = None
            password = None
            if parsed_url.auth:
                auth_parts = parsed_url.auth.split(":", 1)
                if len(auth_parts) < 2:
                    auth_parts += [None]
                username, password = auth_parts
            if not password:
                auth_parts = (
                    # netrc stores things based on entire URL (including username in URL if present)
                    get_netrc_auth(index_url)
                    # keyring could have URL stored by full URL or netloc
                    or get_keyring_auth(parsed_url._replace(auth=None).url, username)
                    or get_keyring_auth(parsed_url.netloc, username)
                )
                if auth_parts is not None:
                    username, password = auth_parts
                if username or password:
                    index_url = parsed_url._replace(auth=f"{username or ''}:{password or ''}").url

            if username and not password:
                logger.info(f"No password found for PyPI index {index_url}")
            elif not username:
                logger.info(f"No username or password found for PyPI index {index_url}")

        authed_index_urls.append(index_url)

    return authed_index_urls


### Code below here was copied from requests.utils to avoid a dependency on requests
NETRC_FILES = (".netrc", "_netrc")


def get_netrc_auth(url, raise_errors=False):
    """Returns the Requests tuple auth for a given url from netrc."""

    netrc_file = os.environ.get("NETRC")
    if netrc_file is not None:
        netrc_locations = (netrc_file,)
    else:
        netrc_locations = (f"~/{f}" for f in NETRC_FILES)

    try:
        from netrc import NetrcParseError, netrc

        netrc_path = None

        for f in netrc_locations:
            try:
                loc = os.path.expanduser(f)
            except KeyError:
                # os.path.expanduser can fail when $HOME is undefined and
                # getpwuid fails. See https://bugs.python.org/issue20164 &
                # https://github.com/psf/requests/issues/1846
                return

            if os.path.exists(loc):
                netrc_path = loc
                break

        # Abort early if there isn't one.
        if netrc_path is None:
            return

        ri = urlparse(url)

        host = ri.netloc.split(":")[0]

        try:
            _netrc = netrc(netrc_path).authenticators(host)
            if _netrc:
                # Return with login / password
                login_i = 0 if _netrc[0] else 1
                return (_netrc[login_i], _netrc[2])
        except (NetrcParseError, OSError):
            # If there was a parsing error or a permissions issue reading the file,
            # we'll just skip netrc auth unless explicitly asked to raise errors.
            if raise_errors:
                raise

    # App Engine hackiness.
    except (ImportError, AttributeError):
        pass


def get_keyring_auth(url: str, username: Optional[str]) -> Optional[Tuple[Optional[str], Optional[str]]]:
    """Returns the Requests tuple auth for a given url from keyring."""

    if not HAVE_KEYRING:
        return None

    if hasattr(keyring, "get_credential"):  # type: ignore[reportPossiblyUnboundVariable]
        logger.debug(f"Getting credentials from keyring for {url}")
        cred = keyring.get_credential(url, username)  # type: ignore[reportPossiblyUnboundVariable]
        if cred is not None:
            return cred.username, cred.password
        return None

    if username is not None:
        logger.debug(f"Getting password from keyring for {url}")
        password = keyring.get_password(url, username)  # type: ignore[reportPossiblyUnboundVariable]
        if password:
            return username, password
    return None


def make_coiled_local_name(dirname: str):
    cleaned_name = re.sub(r"[^\w\d.]+", "_", dirname, flags=re.UNICODE)
    cleaned_name = re.sub(r"_+", "_", cleaned_name)
    if cleaned_name.startswith("_"):
        cleaned_name = cleaned_name.lstrip("_")
    return COILED_LOCAL_PACKAGE_PREFIX + cleaned_name



import datetime
import json
import logging
import os
from typing import Tuple

from distributed import Client

from coiled.utils import supress_logs


class _FakeCredentialsBase:
    def __init__(self, *args, **kwargs): ...


try:
    from google.oauth2.credentials import Credentials as GoogleOAuth2Credentials
except ImportError:
    GoogleOAuth2Credentials = _FakeCredentialsBase


logger = logging.getLogger(__name__)


def get_gcp_local_session_token(scopes=None, set_local_token_env: bool = False) -> dict:
    auth_logger = None
    comp_logger = None
    auth_logger_level = 0
    comp_logger_level = 0

    try:
        import google.auth
        from google.auth.exceptions import RefreshError
        from google.auth.transport.requests import Request

        with supress_logs(["google.auth._default", "google.auth.compute_engine._metadata"]):
            # get token from local auth
            local_creds, project = google.auth.default(scopes=scopes)
            try:
                local_creds.refresh(Request())
            except RefreshError:
                # when local creds are service account, we need to explicitly specify scope(s)
                # or else we get this exception:
                #   RefreshError:
                #   ('invalid_scope: Invalid OAuth scope or ID token audience provided.',
                #    {'error': 'invalid_scope',
                #     'error_description': 'Invalid OAuth scope or ID token audience provided.'})
                local_creds, project = google.auth.default(
                    scopes=[*(scopes or []), "https://www.googleapis.com/auth/cloud-platform"]
                )
                local_creds.refresh(Request())

        if set_local_token_env:
            # set local env var with token, in case `CoiledShippedCredentials` is being used locally
            os.environ["COILED_LOCAL_CLOUDSDK_AUTH_ACCESS_TOKEN"] = local_creds.token

        return {
            "token": local_creds.token,
            "project": project,
            "expiry": local_creds.expiry.replace(tzinfo=datetime.timezone.utc),
        }
    except Exception as e:
        if "default credentials were not found" not in str(e):
            logger.warning(
                "Could not get token from client GCP session. "
                f"This is not a concern unless you're planning to use forwarded GCP credentials on your cluster. "
                f"The error was: {e}"
            )
    finally:
        if auth_logger:
            auth_logger.setLevel(auth_logger_level)
        if comp_logger:
            comp_logger.setLevel(comp_logger_level)

    return {}


def get_application_default_credentials(scopes=None):
    try:
        import google.auth
    except ImportError:
        raise ImportError(
            "Unable to retrieve Google Application Default Credentials because google-auth is not installed."
        ) from None
    local_creds, project = google.auth.default(scopes=scopes)

    return {
        "client_id": local_creds.client_id,
        "client_secret": local_creds.client_secret,
        "quota_project_id": local_creds.quota_project_id or project,
        "refresh_token": local_creds.refresh_token,
        "type": "authorized_user",
    }


def get_long_lived_adc_to_forward(scopes=None):
    to_forward = {}

    # get *local* credentials
    adf = get_application_default_credentials(scopes)
    adf_json = json.dumps(adf)

    project = adf["quota_project_id"]

    default_config = f"""[core]
    project = {project}
    """

    to_forward["env"] = {
        "GS_OAUTH2_CLIENT_ID": adf["client_id"],
        "GS_OAUTH2_CLIENT_SECRET": adf["client_secret"],
        "GS_OAUTH2_REFRESH_TOKEN": adf["refresh_token"],
        "GOOGLE_CLOUD_PROJECT": project,
    }

    to_forward["files"] = {
        "/gcloud-config/application_default_credentials.json": adf_json,
        "/gcloud-config/configurations/config_default": default_config,
    }

    return to_forward


def send_application_default_credentials(cluster, scopes=None, return_creds: bool = False):
    to_forward = get_long_lived_adc_to_forward(scopes)

    def write_adf_files():
        for path, content in to_forward["files"].items():
            abs_path = os.path.expanduser(path)
            os.makedirs(os.path.dirname(abs_path), exist_ok=True)
            with open(abs_path, "w") as f:
                f.write(content)

        os.makedirs(os.path.expanduser("~/.config"), exist_ok=True)
        container_adc_dir = os.path.expanduser("~/.config/gcloud")

        # files are written to /gcloud-config, which is then symlinked to ~/.config/gcloud
        if not os.path.exists(container_adc_dir):
            os.symlink("/gcloud-config", container_adc_dir, target_is_directory=True)
        # otherwise, maybe we should copy the files to both places?

    cluster.send_private_envs(to_forward["env"])
    with Client(cluster) as client:
        client.run_on_scheduler(write_adf_files)
        client.run(write_adf_files)

    cluster._queue_cluster_event("credentials", "Google Application Default Credentials forwarded")
    print(
        "Google Application Default Credentials have been written to a file on your Coiled VM(s).\n"
        "These credentials will potentially be valid until explicitly revoked by running\n"
        "gcloud auth application-default revoke"
    )

    if return_creds:
        return to_forward


class CoiledShippedCredentials(GoogleOAuth2Credentials):  # type: ignore
    def __init__(
        self,
        token=None,
        refresh_token=None,
        id_token=None,
        token_uri=None,
        client_id=None,
        client_secret=None,
        scopes=None,
        default_scopes=None,
        quota_project_id=None,
        expiry=None,
        rapt_token=None,
        refresh_handler=None,
        enable_reauth_refresh=False,
        granted_scopes=None,
    ):
        if GoogleOAuth2Credentials is _FakeCredentialsBase:
            raise ImportError("Unable to create Google Credentials object because google-cloud-iam is not installed.")

        env_token = self.get_shipped_token()
        if token and env_token and token != env_token:
            raise ValueError(
                "Specified Google OAuth2 token does not match "
                "token shipped by Coiled in CLOUDSDK_AUTH_ACCESS_TOKEN.\n"
                "We recommend not specifying a token and using the shipped token."
            )
        if token and not env_token:
            logger.warning(
                "Instantiating credentials with explicit token, no shipped token "
                "found in CLOUDSDK_AUTH_ACCESS_TOKEN. Refresh (which uses CLOUDSDK_AUTH_ACCESS_TOKEN) "
                "is unlikely to work."
            )

        super().__init__(token=env_token, refresh_handler=self.coiled_token_refresh_handler)

    @staticmethod
    def get_shipped_token():
        token = os.environ.get("CLOUDSDK_AUTH_ACCESS_TOKEN")

        if not token:
            # It's not the normal use-case, but Coiled can also set this env on local client machine
            # so that you can use `CoiledShippedCredentials` (with same OAuth2 token) locally and on cluster.
            token = os.environ.get("COILED_LOCAL_CLOUDSDK_AUTH_ACCESS_TOKEN")

        if not token:
            logger.warning("No Google OAuth2 token found, CLOUDSDK_AUTH_ACCESS_TOKEN env var not set")
        return token

    @staticmethod
    def get_token_expiry(token: str) -> datetime.datetime:
        import httpx

        result = httpx.get(f"https://oauth2.googleapis.com/tokeninfo?access_token={token}")
        data = result.json()
        timestamp = int(data["exp"])
        # note that refresh_handler is expected to return naive utc datetime
        expiry = datetime.datetime.utcfromtimestamp(timestamp)
        return expiry

    def coiled_token_refresh_handler(self, request, scopes) -> Tuple[str, datetime.datetime]:
        # this relies on other Coiled mechanisms to have already shipped a non-expired token to the cluster
        token = self.get_shipped_token()

        if not token:
            from google.auth.exceptions import RefreshError

            raise RefreshError(
                "Coiled was unable to find Google OAuth2 token on the cluster. "
                "See https://docs.coiled.io/user_guide/remote-data-access.html#gcp for details about shipping "
                "OAuth2 tokens from the client to the cluster."
            )

        expiry = self.get_token_expiry(token)

        logger.info(f"CoiledShippedCredentials have been refreshed, new expiration is {expiry}")

        return token, expiry


def get_identity(credentials=None):
    import google.auth
    import httpx
    from google.auth.transport.requests import Request

    if not credentials:
        credentials, project = google.auth.default()
    credentials.refresh(Request())

    result = httpx.get(f"https://oauth2.googleapis.com/tokeninfo?access_token={credentials.token}")
    return result.json()



import click
from rich import print

import coiled

from .. import curl
from ..utils import CONTEXT_SETTINGS
from .better_logs import format_log_event
from .utils import find_cluster


@click.command(
    context_settings=CONTEXT_SETTINGS,
)
@click.option(
    "--workspace",
    default=None,
    help="Coiled workspace (uses default workspace if not specified)."
    " Note: --account is deprecated, please use --workspace instead.",
)
@click.option(
    "--cluster",
    default=None,
    help="Cluster for which to show logs, default is most recent",
)
def show_azure_logs(workspace, cluster):
    with coiled.Cloud(workspace) as cloud:
        cluster_info = find_cluster(cloud, cluster or "")
        cluster_id = cluster_info["id"]
        cluster_name = cluster_info["name"]

        print(f"=== Logs for {cluster_name} ({cluster_id}) ===\n")

        print_azure_log_pages(pull_azure_logs(cloud, cluster_id))


def pull_azure_logs(cloud, cluster_id: int):
    url = f"{cloud.server}/api/v2/logs/cluster/{cluster_id}/init"
    response = curl.sync_request(cloud=cloud, url=url, method="get", data={}, json_output=True)

    if response.get("error"):
        print(response["error"])
        return []

    token = response.get("back")
    lines = []

    while token:
        url = f"{cloud.server}/api/v2/logs/cluster/{cluster_id}/page/{token}"
        response = curl.sync_request(cloud=cloud, url=url, method="get", data={}, json_output=True)
        lines.extend(response["lines"])
        token = response.get("back")

    return sorted(lines, key=lambda line: line["timestamp"])


def print_azure_log_pages(lines):
    for line in lines:
        print(
            format_log_event(
                line, instances={}, pretty=True, show_label=True, show_timestamp=True, show_all_timestamps=False
            )
        )



import enum
import logging
import time

import dask.config
from dask.distributed import Client
from dask.utils import parse_timedelta
from distributed.deploy.adaptive import Adaptive

from .utils import COILED_LOGGER_NAME

logger = logging.getLogger(COILED_LOGGER_NAME)


@enum.unique
class CredentialsPreferred(enum.Enum):
    LOCAL = "local"
    # USER = 'user'
    ACCOUNT = "account"  # doesn't work, should be fixed or deprecated/removed someday
    NONE = None


class CoiledAdaptive(Adaptive):
    def __init__(self, cluster, **kwargs):
        super().__init__(cluster, **kwargs)

        update_interval = dask.config.get("coiled.adaptive.update-plan-interval", None)
        self._update_plan_interval = parse_timedelta(update_interval) if update_interval else None  # in seconds
        self._last_plan_update = time.monotonic()

        target_duration = kwargs.get("target_duration")
        if target_duration:
            with Client(self.cluster) as client:
                client.run_on_scheduler(
                    lambda: dask.config.set({"distributed.adaptive.target-duration": target_duration})
                )

    def config_as_string(self):
        return (
            f"Config is maximum: {self.maximum}, minimum: {self.minimum}, wait_count: {self.wait_count}, "
            f"interval: {self.interval}, target_duration: {self.target_duration}."
        )

    async def scale_up(self, n):
        logger.info(f"Adaptive scaling up to {n} workers.")
        if self.cluster:
            if hasattr(self.cluster, "_submit_cluster_event"):
                await self.cluster._submit_cluster_event(
                    "adaptive",
                    f"scale up to {n} workers",
                    extra_data={"n": n},
                )
            await self.cluster.scale_up(
                n,
                reason=f"Adaptive scaling up to {n} workers. "
                f"Previous intended size was {len(self.cluster._plan)}. "
                f"{self.config_as_string()}",
            )

    async def scale_down(self, workers):
        if not workers:
            return
        logger.info(f"Adaptive is removing {len(workers)} workers.")
        if self.cluster:
            if hasattr(self.cluster, "_submit_cluster_event"):
                await self.cluster._submit_cluster_event(
                    "adaptive",
                    f"removing {len(workers)} workers",
                    extra_data={"workers": list(workers)},
                )

            await self.cluster.scale_down(
                workers,
                reason=f"Adaptive removing {len(workers)} workers. "
                f"Previous intended size was {len(self.cluster._plan)}. "
                f"{self.config_as_string()}",
                force_stop=False,
            )

    async def workers_to_close(self, target: int):
        if self.cluster and hasattr(self.cluster, "workers_to_close"):
            # use our logic for determining which workers to close
            # e.g., if there's an extra worker on scheduler VM, don't close that
            return await self.cluster.workers_to_close(target=target)

        # just in case, fall back to base class method
        return await super().workers_to_close(target=target)

    async def recommendations(self, target: int) -> dict:
        if self.cluster and hasattr(self.cluster, "_set_plan_requested"):
            # update current view of workers if running worker on scheduler
            # since we initially include fake worker in the set but this messes things up if we
            # don't remove later when actual worker has appeared.
            if "extra-worker-on-scheduler" in self.cluster.plan:
                await self.cluster._set_plan_requested()
                self._last_plan_update = time.monotonic()

            # update current view of workers at regular interval, so if prior adaptive requests failed to get desired
            # workers (e.g., there were availability limits which caused workers to go to `error` state), adaptive
            # will be aware that those workers failed and can request new workers.
            # the interval will effectively control how often adaptive will retry to get workers.
            elif self._update_plan_interval and time.monotonic() > self._last_plan_update + self._update_plan_interval:
                if hasattr(self.cluster, "_submit_cluster_event"):
                    await self.cluster._submit_cluster_event(
                        "adaptive",
                        "updating plan based on current state of workers",
                        extra_data={
                            "update_interval": self._update_plan_interval,
                            "last_update": time.monotonic() - self._last_plan_update,
                        },
                    )
                await self.cluster._set_plan_requested()
                self._last_plan_update = time.monotonic()

        return await super().recommendations(target=target)



from packaging.version import Version

try:
    import prefect  # type: ignore # noqa
except ImportError as e:
    raise ImportError("The `coiled.prefect` module requires `prefect` to be installed.") from e

from typing import Optional

import dask.config
import pydantic
from prefect.blocks.core import Block  # type: ignore

if Version(prefect.__version__).major < 3 and Version(pydantic.__version__).major >= 2:
    # Prefect docs say this is required for pydantic>=2
    # https://docs.prefect.io/latest/concepts/blocks/#secret-fields
    from pydantic.v1 import Field, SecretStr
else:
    from pydantic import Field, SecretStr


class Credentials(Block):
    _block_type_name = "Coiled Credentials"
    _logo_url = "https://blog.coiled.io/_static/logo.svg"
    _documentation_url = "https://docs.coiled.io/user_guide/labs/prefect.html"
    _code_example = """
    ```python
    from coiled.prefect import Credentials

    Credentials.load("BLOCK_NAME").login()
    ```
    """

    token: SecretStr = Field(default=..., description="Coiled API token.")
    account: Optional[str] = Field(default=None, description="Deprecated, use ``workspace``.")
    workspace: Optional[str] = Field(default=None, description="Coiled workspace to use.")

    def login(self):
        workspace = self.workspace or self.account
        dask.config.set({
            "coiled.token": self.token.get_secret_value(),
            "coiled.account": workspace,
            "coiled.workspace": workspace,
        })



from __future__ import annotations

import datetime
import io
import json
import os
import pathlib
import shlex
from typing import Dict, List, Optional, Sequence, Union

import click
import dask.config
import fabric
import invoke
from dask.utils import parse_timedelta
from paramiko.ed25519key import Ed25519Key
from rich import print
from rich.console import Console
from urllib3.util import parse_url

import coiled
from coiled.credentials.google import send_application_default_credentials
from coiled.spans import span
from coiled.utils import (
    dict_from_key_val_list,
    error_info_for_tracking,
    short_random_string,
    unset_single_thread_defaults,
)
from coiled.v2.cluster import ClusterKwargs
from coiled.v2.widgets.rich import LightRichClusterWidget

from .sync import SYNC_TARGET, start_sync, stop_sync
from .utils import CONTEXT_SETTINGS

# This directory is used by shell scripts that wrap the user's command
# Note that this directory is also referenced in code that runs in our
# cloud_env image, so if it's changed here, that code should also be updated.
COMMAND_DIR = "/scratch/run"
FILE_TEMP_DIR = "/scratch/run-temp-files"

DEFAULT_DASK_CONTAINER = "tmp-dask-1"


class KeepaliveSession:
    def __init__(self, cluster, prefix="", monitor_proc_activity=False):
        self.cluster = cluster
        self.monitor_proc_activity = monitor_proc_activity
        rand_uuid = short_random_string()
        self.session_id = f"{prefix}-{rand_uuid}" if prefix else rand_uuid

    def __enter__(self):
        # keepalive session lets us use keepalive without dask client
        self.cluster._call_scheduler_comm(
            "coiled_add_keepalive_session", name=self.session_id, monitor_proc_activity=self.monitor_proc_activity
        )

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.cluster._call_scheduler_comm("coiled_end_keepalive_session", name=self.session_id)


def get_ssh_connection(cloud, cluster_id) -> fabric.connection.Connection:
    ssh_info = cloud.get_ssh_key(cluster_id=cluster_id)

    with io.StringIO(ssh_info["private_key"]) as f:
        pkey = Ed25519Key(file_obj=f)

    ssh_address = ssh_info["scheduler_address_to_use"]
    if not ssh_address:
        raise coiled.exceptions.CoiledException(
            f"Unable to connect to cluster (id={cluster_id}) because we don't have IP address for it."
        )

    connection = fabric.connection.Connection(ssh_address, user="ubuntu", connect_kwargs={"pkey": pkey})

    return connection


def write_via_ssh(connection, content, path, mode=None):
    with io.StringIO(content) as f:
        connection.put(f, path)
    if mode:
        connection.sftp().chmod(path, mode)


def write_files_into_container(connection, files: Dict[str, str]):
    make_remote_dir(connection, FILE_TEMP_DIR)
    temp_paths = {path: f"{FILE_TEMP_DIR}/{short_random_string()}-{path.split('/')[-1]}" for path in files.keys()}
    # write files to temporary (staging) directory inside /scratch, so they're accessible inside container
    for path, content in files.items():
        write_via_ssh(connection, content=content, path=temp_paths[path], mode=0o755)

    # run commands inside container to move files to desired path
    move_commands = []
    for path, temp_path in temp_paths.items():
        # direct path manipulation because this is linux path, not local os path
        path_dir = "/".join(path.split("/")[:-1])
        move_commands.append(f"mkdir -p {path_dir} && mv {temp_path} {path}")
    for command in move_commands:
        # use `bash -c ...` so that `~` in path expands to user home inside container
        connection.run(command_with_docker_exec(f"bash -c '{command}'"), hide=True, in_stream=False)


def upload_file(connection, f, specified_root=None) -> str:
    cwd = os.path.abspath(os.path.curdir)
    base = os.path.basename(f)
    is_under_cwd = os.path.commonpath((os.path.abspath(f), cwd)) == cwd

    if is_under_cwd:
        relative_to = os.path.curdir
    elif specified_root:
        relative_to = specified_root
    else:
        relative_to = None

    if relative_to:
        # For file that's inside cwd, keep the relative path.
        # Note that this could be different from how you specified the path, for example
        #   cwd=/foo/bar
        #   coiled run --file /foo/bar/nested/path.txt
        # file will be copied to /scratch/nested/path.txt
        # which is a little confusing, but means it's equivalent to
        #   coiled run --file ~/nested/path.txt
        # which does feel natural.

        # For file that's not inside cwd, keep path relative if the user specified a directory for upload.
        # For example, if user specified `--file /absolute/subdir/`, then preserve path structure relative
        # to `/absolute/subdir/`, so `/absolute/subdir/foo/bar.txt` would go to `/scratch/subdir/foo/bar.txt`.
        specified_path_dir = os.path.dirname(os.path.relpath(f, relative_to))
        remote_dir = f"/scratch/{specified_path_dir}/"
        make_remote_dir(connection, remote_dir)
    else:
        remote_dir = "/scratch/"

    connection.put(f, remote_dir)

    # we want path on Linux VM, which might not match os.path.join run client-side, so join path manually
    # remote_dir should already end in "/"
    return f"{remote_dir}{base}"


def run_via_ssh(
    cloud,
    cluster,
    info: dict,
    command: List[str],
    file: List[str],
    env: dict,
    skip_entrypoint: bool,
    interactive: bool,
    detach: bool,
    env_unset: Optional[Sequence[str]] = None,
    workdir: str = "/scratch",
    as_root: bool = False,
    in_container_files: Optional[Dict[str, str]] = None,
    console=None,
):
    console = console or Console()

    connection = get_ssh_connection(cloud, cluster.cluster_id)
    results = None

    original_command = shlex.join(command)
    callstack = [{"code": original_command, "relative_line": 0}]
    # Extract and upload files from `command`
    command = shlex.split(original_command)
    info["files-implicit"] = []  # implicit files are part of the command, e.g., `foo.py` in `coiled run python foo.py`
    for idx, implicit_file in enumerate(command):
        if os.path.exists(implicit_file) and os.path.isfile(implicit_file):
            info["files-implicit"].append(implicit_file)

            # this will preserve path structure relative to cwd
            # so `coiled run python ./subdir/foo.py` will go to `/scratch/subdir/foo.py`
            remote_path = upload_file(connection, implicit_file)

            # adjust command to reference path on VM
            command[idx] = remote_path

            # include code files in the "callstack"
            if implicit_file.endswith((".py", ".sh")):
                try:
                    with open(implicit_file) as f:
                        contents = f.read()
                    callstack.append({"code": contents, "filename": implicit_file})
                except Exception:
                    pass

    # Upload user-specified files too
    info["files-explicit"] = file

    files_to_upload = []
    for f in file:
        path = pathlib.Path(f)
        if not path.exists():
            raise FileNotFoundError(f"Cannot find specified file {f}")

        if path.is_file():
            files_to_upload.append({"f": path})
        elif path.is_dir():
            # for paths outside cwd, parent_dir is used as the root so that path structure from there is preserved
            parent_dir = pathlib.Path(path).parent
            for subfile in path.rglob("*"):
                if subfile.is_file():
                    files_to_upload.append({"f": subfile, "specified_root": parent_dir})

    if files_to_upload:
        console.print(
            f"Uploading {len(files_to_upload)} file{'s' if len(files_to_upload) > 1 else ''} "
            "from local machine to cloud VM..."
        )
    for i, file_to_upload in enumerate(files_to_upload):
        try:
            mb_size = file_to_upload["f"].stat().st_size / 1_000_000
            if mb_size > 1:
                console.print(f"  {file_to_upload['f']} is {mb_size:.2f}MiB, this may be slow to upload")
        except Exception:
            pass
        upload_file(connection, **file_to_upload)
        if i and (i % 20 == 0 or i + 1 == len(files_to_upload)):
            console.print(f"  {i + 1}/{len(files_to_upload)} files uploaded")

    if skip_entrypoint:
        entrypoint = ""
    else:
        entrypoint = get_entrypoint(connection, container_name=DEFAULT_DASK_CONTAINER)
        if "cloud-env" in entrypoint:
            # hack for cloud-env containers (e.g., package sync) currently required to avoid re-downloading senv
            # once container is suitably changed, we should just be able to use entrypoint like any other container
            entrypoint = "micromamba run -p /opt/coiled/env"

    command_string = shlex.join(command)
    if not interactive:
        command_string = wrap_command_for_logs_and_stdout(connection, command_string, original_command)

    container_command = command_with_docker_exec(
        command=command_string,
        interactive=interactive,
        detach=detach,
        env=env,
        env_unset=env_unset,
        as_root=as_root,
        workdir=workdir,
        entrypoint=entrypoint,
    )

    if in_container_files:
        write_files_into_container(connection, files=in_container_files)

    try:
        with KeepaliveSession(cluster=cluster, prefix="ssh", monitor_proc_activity=detach), span(
            cluster, callstack=callstack
        ):
            optional_kwargs = {"in_stream": False} if not interactive else {}
            results = connection.run(container_command, hide=not interactive, pty=interactive, **optional_kwargs)
    except invoke.exceptions.UnexpectedExit as e:
        results = str(e)
    except Exception as e:
        raise e

    return connection, container_command, results


def wrap_command_for_logs_and_stdout(
    connection: fabric.connection.Connection, command_string: str, original_command: str
):
    command_id = f"{datetime.datetime.now(tz=datetime.timezone.utc).isoformat()}_{coiled.utils.short_random_string()}"

    inner_command_path = f"{COMMAND_DIR}/{command_id}.sh"
    tee_command = f"{COMMAND_DIR}/{command_id}-tee.sh"

    make_remote_dir(connection, COMMAND_DIR)

    # write the user's command as a shell file
    write_via_ssh(connection, command_string, inner_command_path, mode=0o755)
    # wrapper file that calls user command, sending output to pseudo-tty and stdout (which goes into logs)
    escaped_original_command = original_command.replace('"', '"')
    tee_script = f"""
    echo "{escaped_original_command}\t{command_id}" >> {COMMAND_DIR}/list
    {inner_command_path} 2>&1 | tee -a /proc/1/fd/1
    """
    write_via_ssh(connection, tee_script, tee_command, mode=0o755)
    command_string = f"bash {tee_command}"

    return command_string


def command_with_docker_exec(
    command: str,
    *,
    container_name: str = DEFAULT_DASK_CONTAINER,
    interactive: bool = False,
    detach: bool = False,
    env: Optional[dict] = None,
    env_unset: Optional[Sequence[str]] = None,
    as_root: bool = False,
    workdir: str = "/scratch",
    entrypoint: str = "",
):
    env = env or {}

    container_user = "-u root" if as_root else ""
    docker_opts = "-it" if interactive else ""
    docker_opts = f"{docker_opts} --detach" if detach else docker_opts

    env_opts = " ".join(f"--env {key}={val}" for key, val in env.items()) if env else ""
    env_unset_opts = " ".join(f"--env {key}" for key in env_unset) if env_unset else ""

    return (
        f"docker exec {container_user} --workdir {workdir} {env_opts} {env_unset_opts} "
        f"{docker_opts} {container_name} {entrypoint} {command}"
    )


def check_explicit_files(files):
    for f in files:
        path = pathlib.Path(f)
        if not path.exists():
            raise FileNotFoundError(f"Cannot find specified file {f}")


def make_remote_dir(connection, dir):
    connection.run(f"mkdir -p {dir} && sudo chmod a+wrx {dir}", in_stream=False)


def get_entrypoint(connection, container_name) -> str:
    entrypoint = ""
    x = connection.run(
        f"docker inspect -f '{{{{json .Config.Entrypoint}}}}' {container_name}", hide=True, in_stream=False
    )
    if x.stdout:
        entrypoint = json.loads(x.stdout)

        if entrypoint:
            # hack so we don't use `tini` for each run, remove everything between `tini` and `--`
            # e.g., "tini -g -- /usr/bin/prepare.sh" -> "/usr/bin/prepare.sh"
            if "tini" in entrypoint[0]:
                if "--" in entrypoint:
                    after_tini = entrypoint.index("--")
                    entrypoint = entrypoint[after_tini + 1 :]

            entrypoint = " ".join(entrypoint) if entrypoint else ""
    return entrypoint or ""


@click.command(context_settings=CONTEXT_SETTINGS)
@click.option(
    "--name",
    default=None,
    help="Run name. This identifier controls whether `coiled run` "
    "invocations are dispatched to the same cloud VM or not. "
    "Use the same name to run multiple commands on the same VM. "
    "Defaults to a unique name with no VM reuse.",
)
@click.option(
    "--account",
    "--workspace",
    default=None,
    help="Coiled workspace (uses default workspace if not specified)."
    " Note: ``--account`` is deprecated, please use ``--workspace`` instead.",
)
@click.option(
    "--software",
    default=None,
    help=(
        "Software environment name to use. If neither software nor container is specified, "
        "all the currently-installed Python packages are replicated on the VM using package sync."
    ),
)
@click.option(
    "--container",
    default=None,
    help=(
        "Container image to use. If neither software nor container is specified, "
        "all the currently-installed Python packages are replicated on the VM using package sync."
    ),
)
@click.option(
    "--vm-type",
    default=[],
    multiple=True,
    help="VM type to use. Specify multiple times to provide multiple options.",
)
@click.option(
    "--gpu",
    default=False,
    is_flag=True,
    help="Have a GPU available.",
)
@click.option(
    "--region",
    default=None,
    help="The cloud provider region in which to run the notebook.",
)
@click.option(
    "--disk-size",
    default=None,
    help="Use larger-than-default disk on VM, specified in GiB.",
)
@click.option(
    "--keepalive",
    default=None,
    help=(
        "Keep your VM running for the specified time, even after your command completes. "
        "In seconds (``--keepalive 60``) unless you specify units (``--keepalive 3m`` for 3 minutes)."
        "Default to shutdown immediately after the command finishes."
    ),
)
@click.option(
    "--file",
    "-f",
    default=[],
    multiple=True,
    help=(
        "Local files required to run command. Can be either individual file or entire directory. "
        "Multiple values can be specified, such as ``--file foo.txt --file my-subdir/``."
    ),
)
@click.option(
    "--env",
    "-e",
    default=[],
    multiple=True,
    help=(
        "Environment variables securely transmitted to run command environment. "
        "Format is ``KEY=val``, multiple vars can be set with separate ``--env`` for each."
    ),
)
@click.option("--subdomain", default=None, help="Custom subdomain for the VM hostname.")
@click.option(
    "--allow-ssh-from",
    default="me",
    type=str,
    help=(
        "IP address or CIDR from which connections to port 22 (SSH) are open; "
        "can also be specified as 'everyone' (0.0.0.0/0) or 'me' (automatically determines public IP detected "
        "for your local client). Note that ``coiled run`` relies on SSH connection for executing commands on VM."
    ),
)
@click.option(
    "--port",
    default=[],
    type=int,
    multiple=True,
    help=(
        "Open extra ports in network firewall for inbound connections "
        "(multiple ports can be set with separate ``--port`` for each)."
    ),
)
@click.option(
    "--interactive",
    "-it",
    default=False,
    is_flag=True,
    help="Open an interactive session, e.g., ``coiled run --interactive bash`` or ``coiled run --interactive python``.",
)
@click.option(
    "--detach", default=False, is_flag=True, help="Start the run in the background, don't wait for the results."
)
@click.option(
    "--sync", default=False, is_flag=True, help="Sync files between local working directory and ``/scratch/synced``."
)
@click.option("--root", default=False, is_flag=True, help="Act as root in Docker container.")
@click.option(
    "--skip-entrypoint",
    default=False,
    is_flag=True,
    hidden=True,
)
@click.option(
    "--no-credential-forwarding",
    default=False,
    is_flag=True,
    help=(
        "Disable automatic forwarding of local cloud credentials; "
        "use if you want your code to use credentials attached to the VM (e.g., AWS Instance Profile)."
    ),
)
@click.option(
    "--forward-gcp-adc",
    default=False,
    is_flag=True,
    help="Forward long-lived Google Cloud Application Default Credentials to VM for data access.",
)
@click.option(
    "--tag",
    "-t",
    default=[],
    multiple=True,
    help="Tags. Format is ``KEY=val``, multiple vars can be set with separate ``--tag`` for each.",
)
@click.option(
    "--sync-ignore",
    default=None,
    multiple=True,
    help="Paths to not sync when using ``--sync``.",
)
@click.option(
    "--mount-bucket",
    default=None,
    multiple=True,
    help="S3 or GCS bucket(s) to mount as volumes.",
)
@click.argument("command", nargs=-1)
def run(
    name: Optional[str],
    account: Optional[str],
    software: Optional[str],
    container: Optional[str],
    vm_type: Sequence[str],
    gpu: bool,
    region: Optional[str],
    disk_size: Optional[int],
    keepalive,
    file,
    env,
    interactive: bool,
    subdomain: Optional[str],
    allow_ssh_from: str,
    port: List[int],
    detach: bool,
    sync: bool,
    root: bool,
    skip_entrypoint: bool,
    no_credential_forwarding: bool,
    forward_gcp_adc: bool,
    tag,
    sync_ignore,
    mount_bucket,
    command,
):
    """
    Run a command on the cloud.
    """
    info = start_run(
        name=name,
        workspace=account,
        software=software,
        container=container,
        vm_type=vm_type,
        gpu=gpu,
        region=region,
        disk_size=disk_size,
        keepalive=keepalive,
        file=file,
        env=env,
        tag=tag,
        interactive=interactive,
        detach=detach,
        skip_entrypoint=skip_entrypoint,
        command=command,
        subdomain=subdomain,
        open_extra_ports=port,
        allow_ssh_from=allow_ssh_from,
        sync=sync,
        root=root,
        no_credential_forwarding=no_credential_forwarding,
        forward_gcp_adc=forward_gcp_adc,
        sync_ignore=sync_ignore,
        mount_bucket=mount_bucket,
    )
    return info["exit_code"]


def start_run(
    command: List[str],
    *,
    name: Optional[str] = None,
    file: Optional[List[str]] = None,
    env: Optional[List[str]] = None,
    tag: Optional[List[str]] = None,
    interactive: bool = False,
    detach: bool = False,
    skip_entrypoint: bool = False,
    workspace: Optional[str] = None,
    software: Optional[str] = None,
    container: Optional[str] = None,
    vm_type: Optional[Sequence[str]] = None,
    gpu: bool = False,
    region: Optional[str] = None,
    disk_size: Optional[int] = None,
    keepalive: Optional[Union[float, str]] = None,
    account: Optional[str] = None,
    cluster_type_tag: str = "run/cli",
    subdomain: Optional[str] = None,
    open_extra_ports: Optional[List[int]] = None,
    allow_ssh_from: str = "me",
    sync: bool = False,
    root: bool = False,
    no_credential_forwarding: bool = False,
    forward_gcp_adc: bool = False,
    sync_ignore: Optional[List[str]] = None,
    mount_bucket: Optional[List[str]] = None,
):
    runtime_env_dict = dict_from_key_val_list(env)
    tags = dict_from_key_val_list(tag)
    file = file or []

    dask_env = unset_single_thread_defaults()
    if container and "rapidsai" in container:
        dask_env = {"DISABLE_JUPYTER": "true", **dask_env}  # needed for "stable" RAPIDS image

    if not command:
        raise ValueError("command must be specified")

    if sync and file:
        raise ValueError("--sync and --file cannot be used together")
    if sync and detach:
        raise ValueError("--sync and --detach cannot be used together")

    if detach and keepalive is None:
        # For --detach, we don't want to shut down cluster on close.
        # We'll use keepalive plugin to monitor the "detached" processes
        # and keep cluster alive while these are running.
        keepalive = "10ms"
    keepalive = parse_timedelta(keepalive)
    shutdown_on_close = keepalive is None

    info = {"command": command, "keepalive": keepalive, "subdomain": subdomain, "exit_code": 0}

    success = True
    exception = None

    # Handle command as string case (e.g. `coiled run "python myscript.py"`)
    if len(command) == 1:
        command = shlex.split(command[0])
    # if user tries `coiled run foo.py` they probably want to run `python foo.py` rather than `foo.py`
    if len(command) == 1 and command[0].endswith(".py"):
        command = ["python", command[0]]

    connection = None
    container_command = None
    results = None

    # fail early if user specified `--file` that doesn't exist
    check_explicit_files(file)

    try:
        with coiled.Cloud(workspace=workspace or account) as cloud:
            workspace = workspace or cloud.default_workspace
            with LightRichClusterWidget(
                workspace=workspace,
                title=f"Running [bold]{shlex.join(command)}[/bold]",
                include_total_cost=False,
                extra_link="..." if open_extra_ports else None,
                extra_link_title="Server",
            ) as widget:
                widget.update(
                    server=cloud.server,
                    cluster_details=None,
                    logs=None,
                    workspace=workspace,
                )
                info["workspace"] = workspace
                cluster_kwargs: ClusterKwargs = {
                    "workspace": workspace,
                    "n_workers": 0,
                    "software": software,
                    "container": container,
                    "scheduler_options": {"idle_timeout": "520 weeks"},  # rely on shutdown_on_close and/or keepalive
                    "scheduler_vm_types": list(vm_type) if vm_type else None,
                    "worker_vm_types": list(vm_type) if vm_type else None,
                    "allow_ssh_from": allow_ssh_from,
                    "open_extra_ports": open_extra_ports,
                    "extra_worker_on_scheduler": True,
                    "environ": dask_env,
                    "scheduler_gpu": gpu,
                    "region": region,
                    "shutdown_on_close": shutdown_on_close,
                    "tags": {**tags, **{"coiled-cluster-type": cluster_type_tag}},
                    "scheduler_disk_size": disk_size,
                    "worker_disk_size": disk_size,
                    "dashboard_custom_subdomain": subdomain,
                    "mount_bucket": mount_bucket,
                }
                cluster_kwargs["name"] = name or f"run-{short_random_string()}"
                cluster_kwargs["cloud"] = cloud
                cluster_kwargs["custom_widget"] = widget  # `Cluster` will update widget until it's "ready"

                if no_credential_forwarding:
                    cluster_kwargs["credentials"] = None
                    dask.config.set({"coiled.use_aws_creds_endpoint": False})
                else:
                    dask.config.set({"coiled.use_aws_creds_endpoint": True})
                with coiled.Cluster(**cluster_kwargs) as cluster:
                    info["cluster_id"] = cluster.cluster_id

                    # set dask config so that cluster started from this cluster will re-use software env
                    runtime_env_dict["DASK_COILED__SOFTWARE"] = cluster._software_environment_name
                    runtime_env_dict["DASK_COILED__SOFTWARE_REQUIRES_VM_ARCH"] = cluster.arch.vm_arch

                    # by default "shutdown on close" is also what enables "keepalive" timeout on clusters,
                    # but we want to use "keepalive" timeout (together with "keepalive sessions") to shut
                    # down cluster after the run (submitted via SSH) is done
                    if not shutdown_on_close:
                        cluster.set_keepalive(keepalive=keepalive)

                    if open_extra_ports:
                        url = (
                            parse_url(cluster._dashboard_address)
                            ._replace(scheme="http", port=open_extra_ports[0], path=None, query=None)
                            .url
                        )
                        widget.update(
                            cluster_details=None,
                            logs=None,
                            extra_link=url,
                        )

                    if interactive:
                        widget.stop()

                    if sync:
                        start_sync(cloud, cluster.cluster_id, console=widget.live.console, ignores=sync_ignore)

                    if forward_gcp_adc:
                        gcp_creds_to_forward = send_application_default_credentials(cluster, return_creds=True)
                        if gcp_creds_to_forward:
                            runtime_env_dict = {**gcp_creds_to_forward.get("env", {}), **runtime_env_dict}

                    connection, container_command, results = run_via_ssh(
                        cloud=cloud,
                        cluster=cluster,
                        info=info,
                        command=command,
                        file=file,
                        interactive=interactive,
                        detach=detach,
                        skip_entrypoint=skip_entrypoint,
                        env=runtime_env_dict,
                        env_unset=list(unset_single_thread_defaults().keys()),
                        workdir=SYNC_TARGET if sync else "/scratch",
                        as_root=root,
                        console=widget.live.console,
                    )

            if sync:
                # this will wait on session flush, so remote files will finish being synced to local system
                stop_sync(cloud, cluster.cluster_id, wait=True)

        if cluster.cluster_id and not interactive:
            if detach:
                print(f"Running [green]{shlex.join(command)}[/green] in background.")
            else:
                print()

                print("Output")
                print("-" * len("Output"))
                print()

            if results:
                if isinstance(results, str):
                    print(results)
                else:
                    print(results.stdout)
                    if results.stderr:
                        print(f"[red]{results.stderr}")

            print()
        return info
    except Exception as e:
        success = False
        info["exit_code"] = 1
        exception = e
        print("[red]There was an unexpected error:[/red]")
        print(exception)
        return info
    finally:
        coiled.add_interaction(
            "coiled-run-cli",
            success=success,
            **info,
            **error_info_for_tracking(exception),
        )



import logging
import os

from distributed.diagnostics.plugin import SchedulerPlugin, WorkerPlugin


class DaskSchedulerWriteFiles(SchedulerPlugin):
    name = "scheduler-write-files"

    def __init__(self, files, symlink_dirs=None):
        self._files_to_write = {**(files or {})}
        self._symlink_dirs = {**(symlink_dirs or {})}

    def start(self, *args, **kwargs):
        files = self._files_to_write
        for path, content in files.items():
            abs_path = os.path.expanduser(path)
            os.makedirs(os.path.dirname(abs_path), exist_ok=True)
            with open(abs_path, "w") as f:
                f.write(content)

        for source_dir, target_dir in self._symlink_dirs.items():
            target_dir = os.path.abspath(os.path.expanduser(target_dir))
            if not os.path.exists(target_dir):
                try:
                    os.symlink(source_dir, target_dir)
                except Exception:
                    logger = logging.getLogger("distributed.scheduler")
                    logger.exception(f"Error creating symlink from {source_dir} to {target_dir}")


class DaskWorkerWriteFiles(WorkerPlugin):
    name = "worker-write-files"

    def __init__(self, files, symlink_dirs=None):
        self._files_to_write = {**(files or {})}
        self._symlink_dirs = {**(symlink_dirs or {})}

    def start(self, *args, **kwargs):
        files = self._files_to_write
        for path, content in files.items():
            abs_path = os.path.expanduser(path)
            os.makedirs(os.path.dirname(abs_path), exist_ok=True)
            with open(abs_path, "w") as f:
                f.write(content)

        for source_dir, target_dir in self._symlink_dirs.items():
            target_dir = os.path.abspath(os.path.expanduser(target_dir))
            if not os.path.exists(target_dir):
                try:
                    os.symlink(source_dir, target_dir)
                except Exception:
                    logger = logging.getLogger("distributed.worker")
                    logger.exception(f"Error creating symlink from {source_dir} to {target_dir}")



from __future__ import annotations

import os
import shlex
import shutil
import subprocess
from typing import List, Optional, Tuple

from rich import print
from rich.console import Console

from .cluster.ssh import add_ssh_access, get_ssh_add_path, get_ssh_path, remove_ssh_access

# Path on VM to sync to.
# We use `/scratch` for now because it's already bind-mounted into docker.
SYNC_TARGET = "/scratch/synced"
MUTAGEN_NAME_FORMAT = "coiled-{cluster_id}"


def can_use_mutagen():
    return get_mutagen_path() and get_ssh_path() and get_ssh_keygen_path()


def start_sync(cloud, cluster_id, console=None, include_vcs=False, debug=False, ignores=None):
    mutagen_path = get_mutagen_path()
    if not mutagen_path:
        return

    console = console or Console()

    _, error = connect_mutagen_sync(
        cloud=cloud, cluster_id=cluster_id, console=console, include_vcs=include_vcs, debug=debug, ignores=ignores
    )
    if error:
        print(error)
    return error


def stop_sync(cloud, cluster_id, message=None, wait=False):
    mutagen_path = get_mutagen_path()
    if cluster_id and mutagen_path and mutagen_session_exists(cluster_id):
        # NOTE: we can't tell if the user asked for `--sync` or not at creation.
        # Best we can do is check if mutagen is installed and the session exists.
        ssh_keygen_path = get_ssh_keygen_path()
        if not ssh_keygen_path:
            return

        # Stop mutagen
        if message:
            print(message)

        if wait:
            subprocess.run([mutagen_path, "sync", "flush", MUTAGEN_NAME_FORMAT.format(cluster_id=cluster_id)])

        subprocess.run([mutagen_path, "sync", "terminate", MUTAGEN_NAME_FORMAT.format(cluster_id=cluster_id)])

        ssh_info = cloud.get_ssh_key(cluster_id)
        scheduler_address = ssh_info["scheduler_hostname"] or ssh_info["scheduler_address_to_use"]

        remove_ssh_access(scheduler_address, key=ssh_info["private_key"])

        # Remove `known_hosts` entries.
        # TODO don't like touching the user's `known_hosts` file like this.
        subprocess.run(
            [
                ssh_keygen_path,
                "-f",
                os.path.expanduser("~/.ssh/known_hosts"),
                "-R",
                scheduler_address,
            ],
            capture_output=True,
        )


def get_mutagen_path() -> Optional[str]:
    mutagen_path = shutil.which("mutagen")
    if not mutagen_path:
        print(
            "[bold red]"
            "mutagen must be installed to synchronize files with notebooks.[/]\n"
            "Install via homebrew (on macOS, Linux, or Windows) with:\n\n"
            "brew install mutagen-io/mutagen/mutagen@0.16\n\n"
            "Or, visit https://github.com/mutagen-io/mutagen/releases/latest to download "
            "a static, pre-compiled binary for your system, and place it anywhere on your $PATH."
        )
        return None
    return mutagen_path


def get_ssh_keygen_path() -> Optional[str]:
    ssh_keygen_path = shutil.which("ssh-keygen")
    if not ssh_keygen_path:
        print("[bold red]Unable to find `ssh-keygen`, you may need to install OpenSSH or add it to your paths.[/]")
        return None
    return ssh_keygen_path


def mutagen_session_exists(cluster_id: int) -> bool:
    mutagen_path = get_mutagen_path()
    if not mutagen_path:
        return False
    sessions = (
        subprocess.run(
            [
                mutagen_path,
                "sync",
                "list",
                "--label-selector",
                f"managed-by=coiled,cluster-id={cluster_id}",
                "--template",
                "{{range .}}{{.Name}}{{end}}",
            ],
            text=True,
            capture_output=True,
        )
        .stdout.strip()
        .splitlines()
    )

    if not sessions:
        return False
    if sessions == [MUTAGEN_NAME_FORMAT.format(cluster_id=cluster_id)]:
        return True

    if len(sessions) == 1:
        raise RuntimeError(
            f"Unexpected mutagen session name {sessions[0]!r}. "
            f"Expected {MUTAGEN_NAME_FORMAT.format(cluster_id=cluster_id)!r}."
        )

    raise RuntimeError(f"Multiple mutagen sessions found for cluster {cluster_id}: {sessions}")


def add_host_to_known_hosts(console, target: str) -> Tuple[bool, str]:
    # Since we may not know the location of known_hosts, use `ssh` to add the key to the agent
    ssh_path = get_ssh_path()
    if not ssh_path:
        raise RuntimeError("ssh not installed")
    result = subprocess.run(
        [
            ssh_path,
            "-o",
            "BatchMode=yes",
            "-o",
            "ConnectTimeout=5",
            "-o",
            "StrictHostKeyChecking=no",
            target,
            "echo",
            "ok",
        ],
        check=False,
        capture_output=True,
        text=True,
    )
    if result.returncode != 0:
        console.print(f"[red]Error attempting to ssh...[/red]\n\n{result.stderr}")
        return False, result.stderr
    return True, ""


def connect_mutagen_sync(
    cloud, cluster_id, console, include_vcs: bool = False, debug=False, ignores: Optional[List[str]] = None
) -> Tuple[bool, str]:
    ignores = ignores or []
    ignores = [*ignores, ".venv"]

    if mutagen_session_exists(cluster_id):
        console.print("[bold]File sync session already active; reusing it.[/]")
    else:
        console.print("[bold]Launching file synchronization...[/]")
        ssh_info = cloud.get_ssh_key(cluster_id)

        scheduler_address = ssh_info["scheduler_hostname"] or ssh_info["scheduler_public_address"]
        target = f"ubuntu@{scheduler_address}"

        add_ssh_access(scheduler_address, key=ssh_info["private_key"])

        added, ssh_error = add_host_to_known_hosts(console, target)
        if not added:
            return False, ssh_error

        if debug:
            ssh_keygen = shutil.which("ssh-keygen")
            if ssh_keygen:
                proc = subprocess.run(
                    [ssh_keygen, "-H", "-F", scheduler_address],
                    check=False,
                    capture_output=True,
                    text=True,
                )
                console.print(f"Existing known_hosts entries for {scheduler_address}:\n[green]{proc.stdout}[/green]\n")
            ssh_add_path = get_ssh_add_path()
            if ssh_add_path:
                proc = subprocess.run(
                    [ssh_add_path, "-l"],
                    check=False,
                    capture_output=True,
                    text=True,
                )
                console.print(f"Existing ssh keys:\n[green]{proc.stdout}[/green]\n")

        # Start mutagen
        sync_command = [
            "mutagen",
            "sync",
            "create",
            "--name",
            MUTAGEN_NAME_FORMAT.format(cluster_id=cluster_id),
            "--label",
            "managed-by=coiled",
            "--label",
            f"cluster-id={cluster_id}",
            "--no-ignore-vcs" if include_vcs else "--ignore-vcs",
            f"--ignore={','.join(ignores)}" if ignores else "",
            "--max-staging-file-size=1 GiB",
            # make files on the remote public (to avoid problems if uid on host != uid in container)
            "--default-file-mode-beta=666",
            "--default-directory-mode-beta=0777",
            ".",
            f"{target}:{SYNC_TARGET}",
        ]
        if debug:
            console.print(f"Sync command:\n[green]{shlex.join(sync_command)}[/green]\n")

        result = subprocess.run(
            sync_command,
            check=False,
            text=True,
            capture_output=True,
        )

        # TODO show output live by wrapping console.print as file object that we pass to subprocess.run (?)
        if result.stderr:
            console.print("[red]Error attempting to connect sync...[/red]\n\n")
            console.out(result.stderr)
        elif result.stdout:
            console.out(result.stdout)

        if result.returncode != 0:
            # there was a problem starting mutagen sync, so no reason to link the sync directory
            # (and most likely ssh is the problem so this too would fail)
            return False, result.stderr

        # From host (outside container), make the directory and set permissions.
        # If container user doesn't match host user, then we can't do this *inside* container.
        outside_dir_setup = f"mkdir -p {SYNC_TARGET} && chmod a+rwx {SYNC_TARGET}"

        # Within the docker container, symlink the sync directory (`/scratch/sync`)
        # into the working directory for Jupyter, so you can actually see the synced
        # files in the Jupyter browser. We use a symlink since the container doesn't
        # have capabilities to make a bind mount.
        # TODO if we don't like the symlink, Coiled could see what the workdir is for
        #  the image before running, and bind-mount `/sync` on the host to `$workdir/sync`
        #  in the container? Custom docker images make this tricky; we can't assume anything
        #  about the directory layout or what the working directory will be.
        inside_dir_setup = f"docker exec tmp-dask-1 bash -c 'ln -s {SYNC_TARGET} .'"

        symlink_command = [
            "ssh",
            target,
            f"{outside_dir_setup} && {inside_dir_setup}",
        ]
        if debug:
            console.print(
                "Symlink command so /sync in container maps to synced directory:\n"
                f"[green]{shlex.join(symlink_command)}[/green]\n"
            )
        subprocess.run(
            symlink_command,
            check=True,
            capture_output=True,
        )
    return True, ""
